"""
ContactAH

author: jbonet
date:   02/2014

@oliva's lab
"""

from . import Contact

class ContactAH(Contact):

    available_distance_types = set(["min", "geometric"])
    atomic_types             = set(["min"])
    description              = 'AminoHetero'

    def __init__(self, aminoacid, heteroatom, threshold_type = "min", threshold_distance = 6):

        self._check_type(requested_type = threshold_type)

        self._threshold_type     = threshold_type
        self._threshold_distance = threshold_distance

        super(ContactAH, self).__init__(residue1 = aminoacid, residue2 = heteroatom)

    #
    # ATTRIBUTES
    #
    @property
    def aminoacid(self):          return self._residue1

    @property
    def heteroatom(self):         return self._residue2

    @property
    def min_distance(self):       return self._distance["min"][2]
    @property
    def min_atoms(self):          return self._distance["min"][0:2]

    @property
    def geometric_distance(self):   return self._distance["geometric"][2]

    #
    # OVERWRITE PARENT FUNCTION
    #
    def _build(self):
        for dist_type in self.available_distance_types:
            self._distance.setdefault(dist_type, None)

        self._distance[self._threshold_type] = self.aminoacid.distance(self.heteroatom, dist_type = self._threshold_type)
        if float(self._distance[self._threshold_type][2]) <= self._threshold_distance and \
           float(self._distance[self._threshold_type][2]) >= 0:
            self._underthreshold = True
            for dist_type in self._distance:
                if dist_type != self._threshold_type:
                    self._distance[dist_type] = self.aminoacid.distance(self.heteroatom, dist_type = dist_type)

    #
    # TOSTRING
    #
    def toString(self, requested_type = "cb"):
        if requested_type != "all":
            self._check_type(requested_type = requested_type)
            ddata = self._distance[requested_type]
            if requested_type in self.atomic_types:
                return '{0.aminoacid.type}:{0.aminoacid.number}:{1[0].name}\t{0.heteroatom.type}:{0.heteroatom.number}:{1[1].name}\t{1[2]:.3f}'.format(self, ddata)
            else:
                return '{0.aminoacid.type}:{0.aminoacid.number}\t{0.heteroatom.type}:{0.heteroatom.number}\t{1[2]:.3f}'.format(self, ddata)
        else:
            return '{0.aminoacid.type}:{0.aminoacid.number}\t{0.heteroatom.type}:{0.heteroatom.number}\t{0.min_distance:.3f}\t{0.geometric_distance:.3f}'.format(self)
        


import unittest, subprocess, gzip, os, shutil

from SBILib.structure import PDB, Complex 
from SBILib.databases import PDBlink
from SBILib.external.blast import BlastExe





class Test_structure(unittest.TestCase):
    def setUp(self):
        self.protein = PDB("Test_pdb1a3q.ent.gz")

    def test_download(self):
        new = PDBlink()
        path = new.get_PDB("1a3q")
        self.assertIsNot(path,False)

    def test_mmcif(self):
        protein = PDB(cif_file = "Test_1a3q.cif.gz")
        self.assertEqual(protein.chain_identifiers, {'A', 'C', 'B', 'D'})
        self.assertEqual(protein.FASTA_format(), '>1A3Q_A\t37 \nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRxxxxxxSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP\n>1A3Q_B\t37 \nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRxxxxxxSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP\n')

    def test_manipulatingProtein(self):
        self.assertEqual(self.protein.chain_identifiers, {'A', 'C', 'B', 'D'})
        protein2 = self.protein.duplicate()
        protein2 = protein2.fuse_chains(["A","B"])
        self.assertEqual(protein2.chain_identifiers , {"A"})
        ChainA = protein2.get_chain_by_id("A")
        ChainE = ChainA.duplicate()
        ChainE.chain = "E"
        protein2 = self.protein.duplicate()
        protein2.add_chain(ChainE)
        self.assertEqual(protein2.chain_identifiers, {'A', 'C', 'B', 'D', 'E'})
        protein3 = PDB()
        deletion = ["E"]
        for chain in protein2.chain_identifiers:
            if chain not in deletion:
                protein3.add_chain(protein2.get_chain_by_id(chain))
                if (protein2.get_chain_by_id(chain).chaintype == "P"):
                    protein3._has_prot = "TRUE"
                elif (protein2.get_chain_by_id(chain).chaintype == "N"):
                    protein3._has_nucl = "TRUE"
        
        self.assertEqual(protein3.chain_identifiers, self.protein.chain_identifiers)


    def test_reading_sequences(self):
        self.maxDiff = None
        ChainA = self.protein.get_chain_by_id("A")
        self.assertEqual(self.protein.FASTA_format(gapped=False), '>TEST_PDB1A3Q_A\t37 \nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP\n>TEST_PDB1A3Q_B\t37 \nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP\n' )
        self.assertEqual(self.protein.FASTA_format(), '>TEST_PDB1A3Q_A\t37 \nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRxxxxxxSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP\n>TEST_PDB1A3Q_B\t37 \nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRxxxxxxSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP\n' )
        self.assertEqual(ChainA.gapped_protein_sequence, 'GPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRxxxxxxSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP')
        self.assertEqual(ChainA.protein_sequence, 'GPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRSLPLKPVISQPIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP' )
        self.assertTrue(self.protein.has_nucleotide)
        first = next(self.protein.nucleotides)
        self.assertEqual(first.globalID, 'TEST_PDB1A3Q_C' )
        self.assertEqual(first.nucleotide_sequence(), 'GGGGAATCCCC' )

class Test_Blast(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.blastresults = None

    def setUp(self):
        url = "ftp://ftp.ebi.ac.uk/pub/databases/uniprot/knowledgebase/uniprot_sprot.fasta.gz"
        output = subprocess.run(["wget", url], capture_output=True, text=True)
        with gzip.open('uniprot_sprot.fasta.gz', 'rb') as f:
            with open('uniprot_sprot.fasta','wb') as f_out:
                shutil.copyfileobj(f,f_out)
        os.remove("uniprot_sprot.fasta.gz")
        self.protein = PDB("Test_pdb1a3q.ent.gz")
        self.chainA = self.protein.get_chain_by_id("A")

    def test_Compile_Blast(self):
        querySequence = self.chainA.protein_sequence
        queryID = self.chainA.globalID
        try:
            blast = BlastExe(database = "uniprot_sprot.fasta")
        except Exception as e:
            self.fail("BlastExe() raised ExceptionType unexpectedly!")
        try:
            self.__class__.blastresults = blast.execute_query_seq(sequenceID = queryID, sequence = querySequence)
        except Exception as e:
            self.fail("blast.execute_query_seq() raised ExceptionType unexpectedly!")
        first = self.__class__.blastresults.get_hits()[0]
        self.assertEqual(first.sequenceID, 'sp|Q00653|NFKB2_HUMAN')
        self.assertTrue(first.evaluate_Rost_twilight_zone())
    
    def test_PIR(self):
        self.maxDiff = None
        self.assertEqual(self.__class__.blastresults.str_PIR(result = 0), '>P1;TEST_PDB1A3Q_A\nsequence:TEST_PDB1A3Q_A:1:.:285:.:.:.:.:.\nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLV\nTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQR\nQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLR------SLPLKPVISQP\nIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDF\nSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP*\n>P1;NFKB2\nstructureX:NFKB2:37:HUMAN:327:HUMAN:.:.:.:.\nGPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLV\nTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQR\nQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRASDGSFSLPLKPVISQP\nIHDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDF\nSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP*'  )
        self.assertEqual(self.__class__.blastresults.get_hits( evalue = 0, tz_type = "ID")[0]._num_seq, 2)
        

class Test_Complex(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.Complex = None
    def setUp(self):
        self.protein = PDB("Test_pdb1a3q.ent.gz")
        self.chainA = self.protein.get_chain_by_id("A")
        

    def test_Complex_calc(self):
        try:
            self.__class__.Complex = Complex(self.protein)
        except Exception as e:
            self.fail("Complex() raised ExceptionType unexpectedly!")
        
    def test_Interactions(self):
        self.assertEqual(len(self.__class__.Complex.PPInterfaces[0]), 100 )
        self.assertEqual( len(self.__class__.Complex.PNInterfaces), 4)
        for chain in self.protein.chains:
            self.assertFalse(chain.has_heteroatoms)
        Hprotein = PDB("Test_pdb1a2w.ent.gz")
        try:
            HComplex = Complex(Hprotein)
        except Exception as e:
            self.fail("Complex() raised ExceptionType unexpectedly!")
        for chain in Hprotein.chains:
            self.assertTrue(chain.has_heteroatoms)



class Test_Loops(unittest.TestCase):
    def setUp(self):
        self.protein = PDB("Test_pdb1a3q.ent.gz")
        self.ChainA = self.protein.get_chain_by_id("A")

    def test_Calcs(self):
        try:
            self.ChainA.calculate_dssp()
        except Exception as e:
            self.fail("calculate_dssp() raised ExceptionType unexpectedly!")
        try:
            self.ChainA.calculate_archs()
        except Exception as e:
            self.fail("calculate_archs() raised ExceptionType unexpectedly!")

        self.assertEqual(self.ChainA.archs[0].aminoacid_sequence, 'YLVIVEQPKQRGFRFRY' )
        self.assertEqual(self.ChainA.archs[0].structure_sequence, 'EEEEEE-B-SSSB--EE' )



class Test_Grafting(unittest.TestCase):
    def setUp(self):
        self.protein = PDB("Test_pdb1a3q.ent.gz")
        self.ChainA = self.protein.get_chain_by_id("A")
        self.protein2 = PDB("Test_pdb2ram.ent.gz")

    def test_compare_loops(self):

        try:
            report = self.protein.compare_loops(self.protein2)
        except Exception as e:
            self.fail("compare_loops() raised ExceptionType unexpectedly!")
        
        self.assertEqual(report[0], 'QueryProt:A,176,182:TargetProt:A,160,166:VISQPIH,VLSHPIF')

    def test_graft(self):
        self.maxDiff = None
        graft = self.protein.graft(self.protein2, 'QueryProt:A,176,182:TargetProt:A,160,166:VISQPIH,VLSHPIF')
        self.assertEqual(graft.get_chain_by_id("A").protein_sequence, 'GPYLVIVEQPKQRGFRFRYGCEGPSHGGLPGASSEKGRKTYPTVKICNYEGPAKIEVDLVTHSDPPRAHAHSLVGKQCSELGICAVSVGPKDMTAQFNNLGVLHVTKKNMMGTMIQKLQRQRLRSRPQGLTEAEQRELEQEAKELKKVMDLSIVRLRFSAFLRSLPLKPVLSHPIFDSKSPGASNLKISRMDKTAGSVRGGDEVYLLCDKVQKDDIEVRFYEDDENGWQAFGDFSPTDVHKQYAIVFRTPPYHKMKIERPVTVFLQLKRKRGGDVSDSKQFTYYP')

        




if __name__ == '__main__':
    unittest.main()




'''
IMPORTANT! Differientatie between sequence position (counts gaps and initial index) and alingment position

Iterables of SeqAli refer to the alignment

jbonet @ oliva's lab 2013
'''
import re
from collections import Counter

from SBILib.error    import SeqAliError as SAE
from SBILib.sequence import Sequence    as Sequence

class SeqAli(object):

	def __init__(self, sequences, sequenceInits, identities = None, positives = None, gaps = None):

		if not isinstance(sequences, list): 	 raise AttributeError('Sequences must be added in a list\n')
		if not isinstance(sequenceInits, list):	 raise AttributeError('Sequence inits must be added in a list\n')
		if len(sequences) != len(sequenceInits): raise AttributeError('One Sequence Init is required for each sequence\n')

		self._seq     = []
		for aliseq in sequences:
			self._seq.append(Sequence(sequence = aliseq))
		self._num_seq = len(self._seq)
		self._segment = []

		self._idx	  = self._set_index(sequenceInits)

		self._segmentation_ok = self._search_segments()

		if not self._segmentation_ok:
			raise SAE(code = 1)

		self._identities = identities
		self._positives  = positives
		self._gaps       = gaps
		self._alipatt    = None
		self._aliptmeth  = None

		self._aligned_aa = self._get_aligned_aa()

	'''ATTRIBUTES'''
	@property
	def sequences(self):			return self._seq
	@property
	def identities(self):			return int(self._identities)
	@property
	def identities_pec(self):		return 100 * float(self._identities)/self._aligned_aa
	@property
	def positives(self):			return int(self._positives)
	@property
	def positives_pec(self):		return 100 * float(self._positives)/self._aligned_aa
	@property
	def gaps(self):					return int(self._gaps)
	@property
	def alignment_pattern(self):	return self._alipatt
	@property
	def is_multiple(self):			return self._num_seq > 2
	@property
	def aligned_aminoacids(self):	return self._aligned_aa
	@property
	def number_of_sequences(self):	return self._num_seq
	@property
	def are_segments_ok(self):		return self._segmentation_ok

	'''INTERNAL GETTERS/SETTERS'''
	def _sequence_position_id(self, refseq, pos): #Only for ungapped (basically to use in _search_segments())
		return self._idx[refseq] + pos - 1

	def _aligned_sequence_length(self, refseq):
		return len(re.sub('-','',self.sequences[refseq].sequence))

	def _sequence_position_from_alignment_position(self, refseq, pos): 
		#Considering over alignment with everything
		#if the alignemnt position is gap, we move up to the next position
		data = self._seq[refseq][:pos]
		c    = 1
		if bool(re.match(r'\S+\-\w\Z',data)): cr = 1
		else:							cr = 0	
		while (data.endswith('-') and (data <= len(self._seq[refseq]))):
			data = self._seq[refseq][:pos + c]
			c += 1
		c = Counter(data)['-']
		if bool(c):
			return self._idx[refseq] + len(data) - Counter(data)['-'] - cr
		else:
			return self._idx[refseq] + pos - 1

	def _alignment_position_from_sequence_position(self, refseq, pos):
		if not self._seq[refseq].is_gapped:
			try: return pos - self._idx[refseq] + 1
			except ValueError as e: raise IndexError(e.message)
		else:
			c = 0
			pregaps = True
			for x in range(len(self._seq[refseq])):
				if pregaps and self._seq[refseq][x] == '-': continue
				pregaps = False
				if pos == self._idx[refseq] + c:
					return  x  + 1
				if x+1 < len(self._seq[refseq]) and self._seq[refseq][x+1] != '-':
					c += 1
			raise IndexError
 
	def _sequence_position_to_sequence_position(self, refseq, destseq, pos):
 		pos = self._alignment_position_from_sequence_position(refseq, pos)
 		return self._sequence_position_from_alignment_position(destseq, pos)

	def _first_segment_value(self, refseq):
		for x in range(len(self._segment[refseq])):
			if self._segment[refseq][x] != '-':
				return self._segment[refseq][x]

	def _last_segment_value(self, refseq):
		for x in reversed(list(range(len(self._segment[refseq])))):
			if self._segment[refseq][x] != '-':
				return self._segment[refseq][x]

	'''FUNCTIONS'''
	def overlap(self, qrefseq, mappedrefseq, otherAli):
		# measured over the SHORTEST sequence
		# ranges 0<->1
		length = min(self._aligned_sequence_length(qrefseq), otherAli._aligned_sequence_length(qrefseq))

		me = self.get_section_from_sequence_position(qrefseq,     
													 self._first_segment_value(qrefseq),     
													 self._last_segment_value(qrefseq))
		he = otherAli.get_section_from_sequence_position(qrefseq, 
														 otherAli._first_segment_value(qrefseq), 
														 otherAli._last_segment_value(qrefseq))

		if me._first_segment_value(mappedrefseq) > he._last_segment_value(mappedrefseq) or\
		   me._last_segment_value(mappedrefseq)  < he._first_segment_value(mappedrefseq):
			return 0

		if me._first_segment_value(mappedrefseq) == he._first_segment_value(mappedrefseq) and\
		   me._last_segment_value(mappedrefseq)  == he._last_segment_value(mappedrefseq):
			return 1

		upper_ini = max(me._first_segment_value(mappedrefseq), he._first_segment_value(mappedrefseq))
		lower_end = min(me._last_segment_value(mappedrefseq),  he._last_segment_value(mappedrefseq))
		cover = lower_end - upper_ini + 1

		return float(cover)/float(length)

	def get_sequence_section_from_sequence_position(self, refseq, ini, end, gapped = True):

		qseq = self.get_section_from_sequence_position(refseq, ini, end)._seq[refseq]

		if gapped: 	return qseq
		else:		
			qseq.do_ungap()
			return qseq

	def get_segment_section_from_sequence_position(self, refseq, ini, end):

		return self.get_section_from_sequence_position(refseq, ini, end)._segment[refseq]

	def increment_sequence_index(self, refseq, new_index):
		self._idx[refseq] += (new_index - 1)
		for x in range(len(self._segment[refseq])):
			self._segment[refseq][x] = self._segment[refseq][x] - 1 + self._idx[refseq]

	def get_respective_coordinate(self, refseq, destseq, pos):
		return self._sequence_position_to_sequence_position(refseq, destseq, pos)

	def get_coverage_of_sequence(self, refseq,  original_seqlength, ini = None, end = None):
		#if a range is given (sequence positions), it only calculates the coverage
		#according to that section (coverage of the section)
		#else, gives the coverage over the full seq length (original_seqlength)
		is_section = (ini is not None or end is not None)
		
		if ini is None: 
			ini = self._first_segment_value(refseq)
		if end is None: 
			end = self._last_segment_value(refseq)

		if ini is None and end is None: # this sequence is all gaps
			return 0

		section = self.get_section_from_sequence_position(refseq, ini, end)

		if is_section:
			original_seqlength = section._aligned_sequence_length(refseq)

		coverage = 0
		tk_seqs = section._tokenize() 
		for x in range(len(tk_seqs[refseq])):
			if bool(tk_seqs[refseq][x]) and sum([int(n[x]) for n in tk_seqs]) > 1:
				coverage += 1

		return float(coverage)/int(original_seqlength)

	def get_section_from_sequence_position(self, refseq, ini, end):
		align_ini = self._alignment_position_from_sequence_position(refseq, ini)
		align_end = self._alignment_position_from_sequence_position(refseq, end)
		if align_ini > len(self) or align_end > len(self): raise IndexError
		if align_ini <= 0 or align_end <= 0:               raise IndexError
		if align_ini == align_end: 
			if align_ini == len(self):
				align_end = None
			else:
				align_end += 1
		return self.__getitem__(slice(align_ini,align_end,1))

	def add_alignment_pattern(self, alipatt, method = 'blast'):
		self._alipatt   = alipatt
		self._aliptmeth = method
		if self._aliptmeth == 'blast' and not self.is_multiple:
			if self._positives is None and self._identities is None and self._gaps is None:
				self._positives  = 0
				self._identities = 0
				self._gaps       = 0
				for x in range(len(self._alipatt)):
					position = self._alipatt[x]
					if position == '+':
						self._positives  += 1
					elif position == ' ':
						if bool(re.search('-',"".join(self[x + 1]))):
							self._gaps += 1
					else:
						self._identities += 1
						self._positives  += 1
		else:
			return NotImplemented

	"""
		FORMULA DEFINITIONS FOR ROST TWILIGHT ZONE:
		Rost, B. (1999). Twilight zone of protein sequence alignments. Protein Engineering, 12(2), 85 94.
		MODIFIED FROM J.Planas-Iglesias IMPLEMENTATION

		INCLUDES 1 public and 2 private functions.
		Rost's Twilight Zone can only be applyed to alignments between TWO (2) SEQUENCES.
		This means is not allowed for MULTIPLE SEQUENCE ALIGNMENT
	"""
	def evaluate_Rost_twilight_zone(self, equation = "ID", parameter = 0):
		if self.is_multiple: raise SAE(code = 2, value = self._num_seq)

		if equation == "ID" :
			isHomolog = self.identities_pec >= self._get_Rost_ID_threshold(param=parameter)
		elif equation == "SIM": 
			isHomolog = self.positives_pec  >= self._get_Rost_SIM_threshold(param=parameter)
		return isHomolog

	def _get_Rost_ID_threshold(self, param=0):
		import decimal, math
		L = self._aligned_aa
		return decimal.Decimal(param)+(480*pow(L,decimal.Decimal('-0.32')*(1+pow(decimal.Decimal(repr(math.e)),decimal.Decimal(repr(float(-L)/1000))))))

	def _get_Rost_SIM_threshold(self, param=0):
		import decimal, math
		L = self._aligned_aa
		return decimal.Decimal(param)+(420*pow(L,decimal.Decimal('-0.335')*(1+pow(decimal.Decimal(repr(math.e)),decimal.Decimal(repr(float(-L)/2000))))))

	'''PRINTING FORMAT'''
	def formatPositions(self, human_readable = False):
		"""
		METHOD: formatPositions()
			 __STRING_GETTER__
			=>Returns:
				self.qpos
				self.hpos
			as a single string in which fragment definitions will be represented as follows:
				qpos[0]:hpos[0],qpos[1]:hpos[1];qpos[2]:hpos[2],qpos[3]:hpos[3]; [...]
				^       ^       ^       ^       ^       ^       ^       ^
				|       |       |       |       |       |       |       |Hit segment 2 end
				|       |       |       |       |       |       |Query segment 2 end
				|       |       |       |       |       |Hit segment 2 start
				|       |       |       |       |Query segment 2 start
				|       |       |       |Hit segment 1 end
				|       |       |Query segment 1 end
				|       |Hit segment 1 start
				|Query segment 1 start

			IN MULTIPLE ALIGNMENTS, a section which is not aligned while other are will display the same ini-start for a
			given segment.
		"""
		outdata = []
		if not human_readable:
			for x in range(len(self._segment[0])):
				text = []
				for i in range(len(self._segment)):
					text.append("{0}".format(self._segment[i][x]))
				outdata.append(":".join(text))
				if x % 2:
					if x != len(self._segment[0]) - 1:
						outdata.append(";")
				else:
					outdata.append(",")
			return "".join(outdata)
		else:
			for i in range(len(self._segment)):
				text = []
				for x in range(len(self._segment[i])):
					text.append("{0:>4}".format(self._segment[i][x]))
					if not x % 2:
						text.append(" :")
					elif x != len(self._segment[0]) - 1:
						text.append(" |")
				text = "".join(text)
				outdata.append(text)
			return "\n".join(outdata)

	'''FUNCTIONS CALLED ON LOAD'''
	def _set_index(self, seqini):
		return seqini

	def _tokenize(self):
		'''
		returns a binary representation of the alignment in an array
		'''
		token_seqs = []
		for seq in self._seq:
			token_seqs.append(seq.tokenize('binary'))
		return token_seqs

	def _get_aligned_aa(self):
		this_len = 0
		tk_seqs = self._tokenize()

		for x in range(len(self._seq[0])):
			profile_str = "".join([n[x] for n in tk_seqs])
			profile_vle = sum([int(i) for i in profile_str])
			if profile_vle > 1:
				this_len += 1
		return this_len

	def _check_segments(self):
		static_length = len(self._segment[0])
		if static_length %2 != 0: return False
		for i in range(len(self._segment)):
			if len(self._segment[i]) != static_length:
				return False
		return True

	def _search_segments(self):
		counter = []
		previous_profile = ''
		for i in range(self._num_seq):
			counter.append(1)
			previous_profile += '0'
			self._segment.append([])

		tk_seqs = self._tokenize()
		for x in range(len(self._seq[0])):
			profile_str = "".join([n[x] for n in tk_seqs])

			for i in range(self._num_seq):
				if x == 0:
					if bool(int(profile_str[i])): self._segment[i].append(self._sequence_position_id(i,counter[i]))
					else:						  self._segment[i].append('-')	
				else:
					if profile_str != previous_profile:
						if bool(int(previous_profile[i])): self._segment[i].append(self._sequence_position_id(i, (counter[i] - 1)))
						else:							   self._segment[i].append('-')
						if bool(int(profile_str[i])): 	   self._segment[i].append(self._sequence_position_id(i, counter[i]))
						else:						  	   self._segment[i].append('-')
					if x == (len(self._seq[0]) - 1):
						if bool(int(profile_str[i])): self._segment[i].append(self._sequence_position_id(i, counter[i]))
						else:						  self._segment[i].append('-')
				if bool(int(profile_str[i])): counter[i] += 1

			previous_profile = profile_str

		x = 0
		while (x < len(self._segment[0])):
			counter = 0
			for y in range(self._num_seq):
				if self._segment[y][x] != '-': counter += 1
			if counter == 1:
				for y in range(self._num_seq):
					self._segment[y].pop(x)
					self._segment[y].pop(x)
				x = x - 2
			x += 2
		if len(self) == 1:
			for y in range(self._num_seq):
				self._segment[y].append(self._segment[y][0])
		return self._check_segments()

	'''OVEWRITTE OBJECT FUNCTIONS'''
	def __len__(self):
		return len(self._seq[0])

	def __getitem__(self, key):
		# We correct the count as one would try to ask for the first position as 1
		try: 		
			int(key)
			if int(key) > len(self): raise IndexError
			return [n[int(key) - 1] for n in self._seq]
		except:
			if not isinstance(key, slice): 	
				raise TypeError()
			else:
				index = key.indices(len(self) + 1)
				if key.stop is not None:
					index = (index[0],index[1]+1,index[2]) #Remember! we go according to align position, not array!!
				else:
					index = (index[0],index[1],index[2])
				sequences = ['' for x in range(self.number_of_sequences)]
				seqInits  = ['' for x in range(self.number_of_sequences)]
				for i in range(*index):
					data = self[i]
					for x in range(len(data)):
						sequences[x] += data[x]
				for x in range(self.number_of_sequences):
					seqInits[x] = self._sequence_position_from_alignment_position(x,index[0])
				newali = self.__class__(sequences, seqInits) 
				if not self._alipatt is None:
					newali.add_alignment_pattern(self._alipatt[index[0]-1:index[1]-1:index[2]],self._aliptmeth)
				return newali
				
	def __iter__(self):
		for x in range(1, len(self) + 1):
			yield self[x]

	def __repr__(self):
		data = []
		for x in range(self._num_seq):
			data.append('{0}\t{1}\t{2}'.format(self._first_segment_value(x), self._seq[x].sequence, self._last_segment_value(x)))
		if not self._alipatt is None:
			data.append('\t{0}'.format(self._alipatt))
		if not self._identities is None and not self._positives is None and not self._gaps is None:
			data.append('I: {0._identities:<4} P: {0._positives:<4} G: {0._gaps:<4}'.format(self))
		data.append(self.formatPositions(True))
		return "\n".join(data)



class Singleton(type):
    '''
    This \href{}{Singleton} design has been obtained from an
    \href{http://stackoverflow.com/q/6760685/2806632}{entry in StackOverflow}.
    '''
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @staticmethod
    def _list():
        for cls in Singleton._instances:
            yield Singleton._instances[cls]



from .MiniRes         import MiniResidue
from SBILib.beans.JSONer import JSONer


class Hetero(MiniResidue, JSONer):
    '''
    {Hetero} includes all the data related to a heteroatom in the PDB.
    This does not include modified AminoAcids or Nucleotides in the main chain.
    '''
    def __init__(self, restype, position):
        super(Hetero, self).__init__(restype, position)
        self._name = ''
        self._form = ''

    @property
    def name(self):
        return self._name.strip()

    @property
    def formula(self):
        return self._form.strip()

    #
    # READ FUNCTIONS
    #
    def add_name(self, line):
        self._name += ' ' + line

    def add_formula(self, line):
        self._form += ' ' + line

    #
    # COMPARATIVE
    #
    def __hash__(self):
        return hash((self._type, self._chain))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return hash(self) == hash(other)
        raise NotImplemented

    #
    # FORMAT FUNCTIONS
    #
    def as_dict(self):
        nobj = {}
        nobj['type']    = self.restype
        nobj['name']    = self.name
        nobj['formula'] = self.formula
        return nobj



import numpy as np
from collections        import Counter
from random             import randint
from ..atom             import AtomOfAminoAcid
from SBILib.external.DSSP  import DSSPExe
from .Arch               import Arch
from .SecondaryStructure import SecondaryStructure


def calculate_dssp(pdb, tmppdb=None, tmpdssp=None, cleanfiles=True):

    prefix     = ".".join([pdb.globalID, str(randint(0, 100))])
    if tmppdb is None:
        tmppdb  = ".".join([prefix, 'pdb'])
    if tmpdssp is None:
        tmpdssp = ".".join([prefix, 'dssp'])

    newchain = pdb.duplicate(hetero=False, water=False)
    newchain.clean()
    fd = open(tmppdb, 'w')
    fd.write(newchain.PDB_format())
    fd.close()

    dsspexe = DSSPExe(tmppdb, tmpdssp, cleanfiles, cleanfiles)
    m = 0
    for n in range(len(pdb.aminoacids)):
        if pdb.aminoacids[n].single_letter == dsspexe.dsspdata[m].aminoacid:
            if pdb.aminoacids[n].has_full_backbone:
                pdb.aminoacids[n].dssp = dsspexe.dsspdata[m]
                step = 1 if m+1 < len(dsspexe.dsspdata) else 0
                m += step
        else:
            pdb.aminoacids[n].dssp = dsspexe.empty_dssp

    if dsspexe.gapped:
        template = list(pdb.gapped_protein_secondary_structure)
        for i in range(len(template)):
            if template[i] == 'x':
                pdb._gaps.append(i)


def locate_valid_secondarystructures(pdb):

    if not pdb.has_dssp:
        pdb.calculate_dssp()

    counter, start_post, reading_type, added = 0, None, 'C', False
    previous_aa = None

    for residue in pdb.aminoacids:
        ss, aa = residue.secondary_structure, residue.identifier
        if not reading_type in SecondaryStructure.structure_regions:
            counter = 0
            if ss in SecondaryStructure.structure_regions:
                start_post, reading_type, counter = aa, ss, 1
        else:
            if ss == reading_type:
                counter += 1
                if counter == SecondaryStructure.min_ss_length[ss]:
                    pdb._ss_valid.append(SecondaryStructure(ss, start_post))
                    added = True
            else:
                if added:
                    pdb._ss_valid[-1]._length = counter
                    pdb._ss_valid[-1]._endp   = previous_aa
                counter, start_post, reading_type, added = 0, aa, ss, False
                if ss in SecondaryStructure.structure_regions:
                    start_post, reading_type, counter = aa, ss, 1

        previous_aa = aa

    if len(pdb._ss_valid) > 0:
        if pdb._ss_valid[-1]._endp is None:
            pdb._ss_valid[-1]._length = counter
            pdb._ss_valid[-1]._endp   = previous_aa

    for ss in pdb._ss_valid:
        ss._struct = pdb.extract(ss._inip, ss._endp)
        ss.calculate_center_of_masses()


def calculate_archs(pdb, limit_internal_ss=100,
                    limit_distance=False, allowed_gaps=0):
    number = 1
    for i in range(len(pdb.secondary_structures)):
        for j in range(i+1, len(pdb.secondary_structures)):
            f1 = pdb._get_structure_array_coordinate(pdb.secondary_structures[i]._inip)
            f2 = pdb._get_structure_array_coordinate(pdb.secondary_structures[i]._endp)
            f3 = pdb._get_structure_array_coordinate(pdb.secondary_structures[j]._inip)
            f4 = pdb._get_structure_array_coordinate(pdb.secondary_structures[j]._endp)

            internal_ss_header = []
            if j-i-1 > limit_internal_ss:
                break
            else:
                ss_count = 2
                intss_types = ''
                for n in range(i+1,j):
                    internal_ss_header.append(pdb.secondary_structures[n].headerformat(ss_count,chr(65 + ss_count - 1)))
                    intss_types += pdb.secondary_structures[n]._sstype
                    ss_count += 1
            newss = pdb.extract(pdb.secondary_structures[i]._inip, pdb.secondary_structures[j]._endp)
            ss_template = Counter(list(newss.gapped_protein_secondary_structure))
            if 'x' in ss_template and ss_template['x'] > allowed_gaps:
                break
            gapcorrectini = 0
            gapcorrectend = 0
            if len(pdb._gaps) > 0:
                if f1 >= pdb._gaps[0]:
                    tmp_str       = pdb.extract(pdb.aminoacids[0].identifier, pdb.secondary_structures[i]._inip)
                    ss_template   = Counter(list(tmp_str.gapped_protein_secondary_structure))
                    gapcorrectini = ss_template['x']
                    if allowed_gaps == 0:
                        gapcorrectend = gapcorrectini
                    else:
                        tmp_str       = pdb.extract(pdb.aminoacids[0].identifier, pdb.secondary_structures[j]._endp)
                        ss_template   = Counter(list(tmp_str.gapped_protein_secondary_structure))
                        gapcorrectend = ss_template['x']

            secondstructurepair = Arch(pdb.globalID,
                                       pdb.secondary_structures[i],
                                       pdb.secondary_structures[j],
                                       j-i-1, f3 - f2 - 1, newss,
                                       pdb._torsionsCA[f1+gapcorrectini:f4+1+gapcorrectend],
                                       pdb._psiphi[(f1+gapcorrectini)*3:(f4+gapcorrectend)*3+6],
                                       number)
            #distance restriction only aplies to archs with internal structures
            if secondstructurepair.is_superarch and \
               (limit_distance and secondstructurepair.cartesian_distance > limit_distance):
                break

            secondstructurepair._inttxt = internal_ss_header
            secondstructurepair._inttyp = intss_types
            if secondstructurepair.is_superarch: pdb._superarchs.append(secondstructurepair)
            else:                                pdb._archs.append(secondstructurepair)
            number += 1

def calculate_torsionsCA(pdb, accuracy = 0):
    for i in range(0,len(pdb.aminoacids) - 3):
        aa0 = pdb.aminoacids[i]
        aa1 = pdb.aminoacids[i+1]
        aa2 = pdb.aminoacids[i+2]
        aa3 = pdb.aminoacids[i+3]

        ca0 = aa0.ca
        ca1 = aa1.ca
        ca2 = aa2.ca
        ca3 = aa3.ca

        if not isinstance(ca0, AtomOfAminoAcid) or not isinstance(ca1, AtomOfAminoAcid) or \
           not isinstance(ca2, AtomOfAminoAcid) or not isinstance(ca3, AtomOfAminoAcid):
            pdb._torsionsCA += '-'
            continue

        d01 = ca0.distance(ca1)
        v01 = np.divide(np.subtract(ca1.coordinates, ca0.coordinates), d01)
        d12 = ca1.distance(ca2)
        v12 = np.divide(np.subtract(ca2.coordinates, ca1.coordinates), d12)
        d23 = ca2.distance(ca3)
        v23 = np.divide(np.subtract(ca3.coordinates, ca2.coordinates), d23)

        c4  = -1 * np.sum(np.multiply(v01,v12))
        s4  = np.sqrt(1 - np.power(c4, 2))
        a4  = np.degrees(np.arctan(s4/c4))
        if a4 < 0: a4 += 180
        c5  = -1 * np.sum(np.multiply(v12,v23))
        s5  = np.sqrt(1 - np.power(c5, 2))
        a5  = np.degrees(np.arctan(s5/c5))
        if a5 < 0: a5 += 180

        u   = (v01[1]*v12[2])-(v01[2]*v12[1]), (v01[2]*v12[0])-(v01[0]*v12[2]), (v01[0]*v12[1])-(v01[1]*v12[0])
        v   = (v12[1]*v23[2])-(v12[2]*v23[1]), (v12[2]*v23[0])-(v12[0]*v23[2]), (v12[0]*v23[1])-(v12[1]*v23[0])
        c6  = np.sum(np.multiply(u,v))/(s4*s5)
        s6  = np.sum(np.multiply(v01,v))/(s4*s5)
        a6  = np.degrees(np.arctan(s6/c6))
        a7  = a6
        if a7 < 0 and s6 > 0: a6+=180
        if a7 > 0 and c6 < 0: a6-=180
        if a6 < 0:            a6+=360

        accuracy_desc = {'4': ['A','B','C','D'],
                         '12':['A','B','C','D','E','F','G','H','I','J','K','L']}
        if accuracy in accuracy_desc:
            number_of_parts = len(accuracy_desc[str(accuracy)])
            for j in range(number_of_parts):
                if a6 >= j*(360/number_of_parts) and a6 < (j+1)*(360/number_of_parts):
                    pdb._torsionsCA += accuracy_desc[str(accuracy)][j]
                    break
                if a6 == 360:
                    pdb._torsionsCA += accuracy_desc[str(accuracy)][-1]
        else:
            if a6 >= 0   and a6 < 90:   pdb._torsionsCA += 'A'
            if a6 >= 90  and a6 < 160:  pdb._torsionsCA += 'B'
            if a6 >= 160 and a6 < 260:  pdb._torsionsCA += 'C'
            if a6 >= 260 and a6 <= 360: pdb._torsionsCA += 'D'

    pdb._torsionsCA += '-'

    if len(pdb._gaps) > 0:
        template = list(pdb.gapped_protein_secondary_structure)
        tomodify = list(pdb._torsionsCA)
        for x in range(len(template)):
            if template[x] == 'x' and template[x-1] != 'x':
                tomodify[x-1] = '-'
                tomodify[x]   = '-'
            elif template[x-1] == 'x':
                if x < len(tomodify):
                    tomodify[x]   = '-'
                tomodify.insert(x+1, '-')
        pdb._torsionsCA = "".join(tomodify)

def calculate_psiphi(pdb):
    step  = 0
    phi   = [] #step 0 = C(n)   - N(n+1) - CA(n+1) - C(n+1)  : phi
    phi.append(0)
    psi   = 0  #step 1 = N(n-1) - CA(n)  - C(n)    - N(n)    : psi
    omega = [] #step 2 = CA(n)  - C(n)   - N(n+1)  - CA(n+1) : omega
    omega.append(0)
    m     = 0
    uncalculable = 0

    for i in range(3*(len(pdb.aminoacids)-1)):
        if step == 0:
            at0 = pdb.aminoacids[m].c
            at1 = pdb.aminoacids[m+1].n
            at2 = pdb.aminoacids[m+1].ca
            at3 = pdb.aminoacids[m+1].c
        elif step == 1:
            at0 = pdb.aminoacids[m].n
            at1 = pdb.aminoacids[m].ca
            at2 = pdb.aminoacids[m].c
            at3 = pdb.aminoacids[m+1].n
        else:
            at0 = pdb.aminoacids[m].ca
            at1 = pdb.aminoacids[m].c
            at2 = pdb.aminoacids[m+1].n
            at3 = pdb.aminoacids[m+1].ca

        if not isinstance(at0, AtomOfAminoAcid) or not isinstance(at1, AtomOfAminoAcid) or \
           not isinstance(at2, AtomOfAminoAcid) or not isinstance(at3, AtomOfAminoAcid) or \
           (step == 2 and uncalculable > 0):
            if step < 2:
                uncalculable += 1
            else:
                if uncalculable > 0:
                    pdb._psiphi += '---'
                    uncalculable = 0
            step += 1
            if step == 3:
                step = 0
                m   += 1
            continue

        d01 = at0.distance(at1)
        v01 = np.divide(np.subtract(at1.coordinates, at0.coordinates), d01)
        d12 = at1.distance(at2)
        v12 = np.divide(np.subtract(at2.coordinates, at1.coordinates), d12)
        d23 = at2.distance(at3)
        v23 = np.divide(np.subtract(at3.coordinates, at2.coordinates), d23)

        c4  = -1 * np.sum(np.multiply(v01,v12))
        s4  = np.sqrt(1 - np.power(c4, 2))
        c5  = -1 * np.sum(np.multiply(v12,v23))
        s5  = np.sqrt(1 - np.power(c5, 2))

        u   = (v01[1]*v12[2])-(v01[2]*v12[1]), (v01[2]*v12[0])-(v01[0]*v12[2]), (v01[0]*v12[1])-(v01[1]*v12[0])
        v   = (v12[1]*v23[2])-(v12[2]*v23[1]), (v12[2]*v23[0])-(v12[0]*v23[2]), (v12[0]*v23[1])-(v12[1]*v23[0])
        c6  = np.sum(np.multiply(u,v))/(s4*s5)
        s6  = np.sum(np.multiply(v01,v))/(s4*s5)
        a6  = np.degrees(np.arctan(s6/c6))
        a7  = a6
        if a7 < 0 and s6 > 0: a6+=180
        if a7 > 0 and c6 < 0: a6-=180

        if step == 0: phi.append(a6)
        if step == 1: psi   = a6
        if step == 2:
            omega.append(a6)
            if omega[-2] >= -60 and omega[-2] <= 60 and len(omega) > 2:
                pdb._psiphi += 'CIS' #cys-proline
            else:
                if phi[-2] >= -180 and phi[-2] < -20:
                    if phi[-2] >= -180 and phi[-2] < -140:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'E11'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'E12'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'I13'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'N14'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'N15'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'N16'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'E17'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'E18'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'E19'
                    elif phi[-2] >= -140 and phi[-2] < -100:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'B21'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'F22'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'I23'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'I24'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'T25'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'T26'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'B27'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'B28'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'B29'
                    elif phi[-2] >= -100 and phi[-2] < -60:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'B31'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'F32'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'H33'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'H34'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'T35'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'T36'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'B37'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'B38'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'B39'
                    elif phi[-2] >= -60 and phi[-2] < -20:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'P41'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'F42'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'H43'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'H44'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'T45'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'T46'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'P47'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'P48'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'P49'
                elif phi[-2] >= 20 and phi[-2] <= 180:
                    if phi[-2] >= 20 and phi[-2] < 60:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'G61'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'G62'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'G63'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'U64'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'U65'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'L66'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'L67'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'M68'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'M69'
                    elif phi[-2] >= 60 and phi[-2] < 100:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'G71'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'G72'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'G73'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'U74'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'U75'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'L76'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'L77'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'M78'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'M79'
                    elif phi[-2] >= 100 and phi[-2] < 140:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'G81'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'G82'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'G83'
                        elif psi >= -60  and psi <  -20: pdb._psiphi += 'U84'
                        elif psi >= -20  and psi <   20: pdb._psiphi += 'U85'
                        elif psi >= 20   and psi <   60: pdb._psiphi += 'S86'
                        elif psi >= 60   and psi <  100: pdb._psiphi += 'S87'
                        elif psi >= 100  and psi <  140: pdb._psiphi += 'S88'
                        elif psi >=140   and psi <= 180: pdb._psiphi += 'S89'
                    elif phi[-2] >= 140 and phi[-2] <= 180:
                        if   psi >= -180 and psi < -140: pdb._psiphi += 'E91'
                        elif psi >= -140 and psi < -100: pdb._psiphi += 'E92'
                        elif psi >= -100 and psi <  -60: pdb._psiphi += 'I93'
                        elif psi >=  -60 and psi <  -20: pdb._psiphi += 'N94'
                        elif psi >=  -20 and psi <   20: pdb._psiphi += 'N95'
                        elif psi >=   20 and psi <   60: pdb._psiphi += 'N96'
                        elif psi >=   60 and psi <  100: pdb._psiphi += 'E97'
                        elif psi >=  100 and psi <  140: pdb._psiphi += 'E98'
                        elif psi >=  140 and psi <= 180: pdb._psiphi += 'E99'
                elif phi[-2] >= -20 and phi[-2] <20:
                    if   psi >= -180 and psi < -140: pdb._psiphi += 'O51'
                    elif psi >= -140 and psi < -100: pdb._psiphi += 'O52'
                    elif psi >= -100 and psi <  -60: pdb._psiphi += '*53'
                    elif psi >=  -60 and psi <  -20: pdb._psiphi += '*54'
                    elif psi >=  -20 and psi <   20: pdb._psiphi += '*55'
                    elif psi >=   20 and psi <   60: pdb._psiphi += '*56'
                    elif psi >=   60 and psi <  100: pdb._psiphi += '*57'
                    elif psi >=  100 and psi <  140: pdb._psiphi += 'O58'
                    elif psi >=  140 and psi <= 180: pdb._psiphi += 'O59'
        step += 1
        if step == 3:
            step = 0
            m   += 1

    pdb._psiphi += '---'

    if len(pdb._gaps) > 0:
        template = list(pdb.gapped_protein_secondary_structure)
        tomodify = [pdb._psiphi[i:i+3] for i in range(0, len(pdb._psiphi), 3)]
        for x in range(len(template)):
            if template[x] == 'x' and template[x-1] != 'x':
                tomodify[x-1] = '---'
                tomodify.insert(x, '---')
            elif template[x] == 'x':
                tomodify.insert(x, '---')
        pdb._psiphi = "".join(tomodify)



import re
from .MiniRes import MiniResidue
from SBILib.beans.JSONer import JSONer

class Site(JSONer):

    def __init__(self, identifier):
        self._id    = identifier
        self._desc  = ''
        self._evid  = ''
        self._spec  = []
        self._bind  = []

    #
    # ATTRIBUTES
    #
    @property
    def identifier(self):
        return self._id

    @property
    def description(self):
        return self._desc

    @property
    def evidence(self):
        return self._evid

    @property
    def residues(self):
        return self._spec

    @property
    def binding(self):
        return self._bind

    @property
    def chains(self):
        data = []
        if len(self.residues) != 0:
            data.extend(self.residues)
        if len(self.binding) != 0:
            data.extend(self.binding)
        if len(data) > 0:
            data = list(set([x.chain for x in data]))
        return data

    #
    # READ FUNCTIONS
    #
    def add_remark(self, line):
        if line.startswith('EVIDENCE_CODE:'):
            self._evid = line.split(':')[1].strip()
        elif line.startswith('SITE '):
            pass
        elif line.startswith('SITE_DESCRIPTION:'):
            self._add_description(line.split(':')[1].strip())
        elif len(line) > 0:
            self._add_description(line.strip())

    def add_residues(self, residues):
        for r in residues:
            if r[0] != 'HOH':  # SKIP WATERS
                self._spec.append(MiniResidue(r[0], r[1]))

    def _add_description(self, line):
        self._desc += line

        m = re.search('BINDING SITE FOR RESIDUE\s+(\w+)', self._desc)
        if m:
            res = m.group(1)
            pos = re.sub('BINDING SITE FOR RESIDUE\s+\w+', '', line)
            self._bind.append(MiniResidue(res, pos.strip().strip(',')))

    #
    # OUT FUNCTION
    #
    def as_dict(self):
        nobj = {}
        nobj['id']          = self.identifier
        nobj['description'] = self.description
        nobj['evidence']    = self.evidence

        if len(self.binding) > 0:
            nobj['binds'] = [x.as_dict() for x in self.binding]
        else:
            nobj['binds'] = self.binding

        if len(self.residues) > 0:
            nobj['residues'] = [x.as_dict() for x in self.residues]
        else:
            nobj['residues'] = []

        return nobj

    def __repr__(self):
        return self.json()



# -*-
#
# @author: jaumebonet
# @email:  jaume.bonet@gmail.com
# @url:    jaumebonet.github.io
#
# @date:   2016-02-12 13:20:21
#
# @last modified by:   jaumebonet
# @last modified time: 2016-02-15 10:59:52
#
# -*-
from pynion import Singleton

element_list = [
    ('H',  'Hydrogen'),    ('He', 'Helium'),       ('Li', 'Lithium'),      ('Be', 'Beryllium'),
    ('B',  'Boron'),       ('C',  'Carbon'),       ('N',  'Nitrogen'),     ('O',  'Oxygen'),
    ('F',  'Fluorine'),    ('Ne', 'Neon'),         ('Na', 'Sodium'),       ('Mg', 'Magnesium'),
    ('Al', 'Aluminium'),   ('Si', 'Silicon'),      ('P',  'Phosphorus'),   ('S',  'Sulfur'),
    ('Cl', 'Chlorine'),    ('Ar', 'Argon'),        ('K',  'Potassium'),    ('Ca', 'Calcium'),
    ('Sc', 'Scandium'),    ('Ti', 'Titanium'),     ('V',  'Vanadium'),     ('Cr', 'Chromium'),
    ('Mn', 'Manganese'),   ('Fe', 'Iron'),         ('Co', 'Cobalt'),       ('Ni', 'Nickel'),
    ('Cu', 'Copper'),      ('Zn', 'Zinc'),         ('Ga', 'Gallium'),      ('Ge', 'Germanium'),
    ('As', 'Arsenic'),     ('Se', 'Selenium'),     ('Br', 'Bromine'),      ('Kr', 'Krypton'),
    ('Rb', 'Rubidium'),    ('Sr', 'Strontium'),    ('Y',  'Yttrium'),      ('Zr', 'Zirconium'),
    ('Nb', 'Niobium'),     ('Mo', 'Molybdenum'),   ('Tc', 'Technetium'),   ('Ru', 'Ruthenium'),
    ('Rh', 'Rhodium'),     ('Pd', 'Palladium'),    ('Ag', 'Silver'),       ('Cd', 'Cadmium'),
    ('In', 'Indium'),      ('Sn', 'Tin'),          ('Sb', 'Antimony'),     ('Te', 'Tellurium'),
    ('I',  'Iodine'),      ('Xe', 'Xenon'),        ('Cs', 'Caesium'),      ('Ba', 'Barium'),
    ('La', 'Lanthanum'),   ('Ce', 'Cerium'),       ('Pr', 'Praseodymium'), ('Nd', 'Neodymium'),
    ('Pm', 'Promethium'),  ('Sm', 'Samarium'),     ('Eu', 'Europium'),     ('Gd', 'Gadolinium'),
    ('Tb', 'Terbium'),     ('Dy', 'Dysprosium'),   ('Ho', 'Holmium'),      ('Er', 'Erbium'),
    ('Tm', 'Thulium'),     ('Yb', 'Ytterbium'),    ('Lu', 'Lutetium'),     ('Hf', 'Hafnium'),
    ('Ta', 'Tantalum'),    ('W',  'Tungsten'),     ('Re', 'Rhenium'),      ('Os', 'Osmium'),
    ('Ir', 'Iridium'),     ('Pt', 'Platinum'),     ('Au', 'Gold'),         ('Hg', 'Mercury'),
    ('Tl', 'Thallium'),    ('Pb', 'Lead'),         ('Bi', 'Bismuth'),      ('Po', 'Polonium'),
    ('At', 'Astatine'),    ('Rn', 'Radon'),        ('Fr', 'Francium'),     ('Ra', 'Radium'),
    ('Ac', 'Actinium'),    ('Th', 'Thorium'),      ('Pa', 'Protactinium'), ('U',  'Uranium'),
    ('Np', 'Neptunium'),   ('Pu', 'Plutonium'),    ('Am', 'Americium'),    ('Cm', 'Curium'),
    ('Bk', 'Berkelium'),   ('Cf', 'Californium'),  ('Es', 'Einsteinium'),  ('Fm', 'Fermium'),
    ('Md', 'Mendelevium'), ('No', 'Nobelium'),     ('Lr', 'Lawrencium'),   ('Rf', 'Rutherfordium'),
    ('Db', 'Dubnium'),     ('Sg', 'Seaborgium'),   ('Bh', 'Bohrium'),      ('Hs', 'Hassium'),
    ('Mt', 'Meitnerium'),  ('Ds', 'Darmstadtium'), ('Rg', 'Roentgenium'),  ('Cn', 'Copernicium')
]


class Element(object):
    def __init__(self, number, symbol, name):
        self.number = int(number)
        self.symbol = symbol
        self.name   = name

    def __str__(self):
        return "{0.number:>3}\t{0.symbol:>2}\t{0.name}".format(self)


class PeriodicTable(object):
    __metaclass__ = Singleton

    def __init__(self):
        self._table = {}
        for i in range(len(element_list)):
            e = element_list[i]
            self._table.setdefault(e[0], Element(i + 1, e[0], e[1]))

    def __contains__(self, key):
        if isinstance(key, str) and key.isdigit():
            key = int(key)
        if isinstance( key, ( int, long ) ):
            key = key - 1
            if key > len(element_list) - 1 or key < 0:
                return False
            else:
                return True
        elif isinstance(key, str):
            return key.capitalize() in self._table
        else:
            return False

    def __getitem__(self, key):
        if isinstance(key, str) and key.isdigit():
            key = int(key)
        if isinstance( key, ( int, long ) ):
            key = key - 1
            if key > len(element_list) - 1 or key < 0:
                return KeyError('No element with this size is stored.')
            else:
                return self._table[element_list[key][0]]
        elif isinstance(key, str):
            return self._table[key.capitalize()]
        else:
            raise KeyError('KeyType should be integer or string')



class BlastError(Exception):
    """
    Complementary Error Class: BlastError   -> raises when there are no hits for a given query
                                            -> raises when there is no path for blastpgp
                                            -> raises when there is an error in blast execution
    """
    def __init__(self, code = None, value = None):

        self.value = value
        self.code  = code

        if self.get_code() is not None:
            self._determine_message()

    def get_value(self):        return self.value
    def get_code(self):         return self.code

    def _determine_message(self):

        if self.get_code()   ==  0:
            self.value = "There is something wrong with the XML blast output file %s\n" %self.get_value()

        elif self.get_code() == -1:
            self.value = "blast execution failed. Returns error:\n\n%s" %self.get_value()

        elif self.get_code() == -2:
            self.value = "blast is not in $PATH enviroment variable.\n"
            self.value = self.get_value() + "The path can be added through the 'executablePath' option.\n"

        elif self.get_code() == -3:
            self.value = "Input fasta file %s does not exist.\n" %self.get_value()

        elif self.get_code() == -4:
            self.value = "Input fasta file %s is a multi-fasta file\n" %self.get_value() 

        elif self.get_code() == -5:
            self.value = "Query database %s does not exist\n" %self.get_value()

        elif self.get_code() == -6:
            if len(self.get_value()) == 3:
                text = "In order to perform a blastpgp search, the database MUST be formated (see blast's user guide).\n"
                self.value = text
            else:
                text = "Some of the format files for the database generated with formatdb are missing.\n"
                text = text + "Please check %s" %(",".join(self.get_value()))
                self.value = text

        elif self.get_code() == -7:
            self.value = "The specifyed weight matrix is not supported.\nSupported matrices are: %s\n" %self.get_value()

        elif self.get_code() == -8:
            self.value = "For the given matrix only certain gap opening and extension values are allowed:\n %s\n" %self.get_value()

        elif self.get_code() == -9:
            self.value = "The given path %s does not lead to the blastpgp executable.\n" %self.get_value()

        elif self.get_code() == -10:
            self.value = "The kind of blast can only be 'prot' or 'nucl'."

        elif self.get_code() == -11:
            pass
        
        elif self.get_code() ==  1:
            pass

        elif self.get_code() ==  2:
            self.value = "The BlastHandler has been loaded with a blast XML output file, no need to execute blastpgp\n"

        elif self.get_code() == 3:
            pass

    def __str__(self):
        error_str = "\n\n ###################\n[BlastError Type: %d]\n ###################\n%s\n" %(self.get_code(), self.get_value())
        if self.get_code() < 0:
            error_str = error_str + "\nBlastErrors with negative type indicate that the blast search has not been executed.\n"
        elif self.get_code() > 0:
            error_str = error_str + "\nBlastErrors with positive type are related to problems during the parsing of the blast XML output files.\n"
        return error_str


import unittest
from test_modules import * 


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(Test_structure))
    suite.addTest(unittest.makeSuite(Test_Blast))
    suite.addTest(unittest.makeSuite(Test_Complex))
    suite.addTest(unittest.makeSuite(Test_Loops))
    suite.addTest(unittest.makeSuite(Test_Grafting))
    runner = unittest.TextTestRunner()
    runner.run(suite)



'''
Check BLAST user's guide at: http://www.ncbi.nlm.nih.gov/books/NBK1763/

jbonet @ boliva's lab 2013
'''
import os, sys, time, copy, re, configparser

from SBILib.beans.Executable    import Executable
from SBILib.beans.File     import File
from SBILib.sequence  import Fasta      as Fasta
from SBILib.error     import BlastError as BE
from SBILib           import SBIglobals

from SBILib.external.blast import BlastResult as BR
from SBILib.external.blast import BlastHit    as BH

global default_configuration_file
default_configuration_file = os.path.join(os.path.normpath(os.path.join(os.path.dirname(__file__),'..')),'configSBI.txt')

class BlastExe(object):

    def __init__(self, database, search_type = 'prot'):

        #Search Type Check
        if search_type not in set(['prot','nucl']):
            raise BE(-10)
        self._search_type = search_type

        #Blast executable configuration
        self._configurator = configparser.RawConfigParser(allow_no_value=True)
        self._configurator.read(os.getenv('SBI_CONFIG_FILE',default_configuration_file))
        self._exe    = Executable(executable    = self._configurator.get('blast','executable'),
                                  path          = self._configurator.get('blast','path'),
                                  variable_path = self._configurator.get('blast','variable_path'))

        #Database Configuration
        self._database = self._check_database(os.path.abspath(database))
        if os.path.isfile(self._database.file.full + ".idx"):
            self._idx = File(file_name = self._database.file.full + ".idx", action = 'r')
        else:
            self._idx = None

        #Adding fixed blast parameters
        self._exe.add_attribute(self._database.file.full, '-db')
        self._exe.add_attribute('5', '-outfmt')
        self._exe.add_parameter('-lcase_masking')

        SBIglobals.alert('debug', self, 'New Blast Executable created.\nBlast executable at {0}\n'.format(self._exe.full_executable))

        self._selfHit     = False
        self._hitIDformat = 'single'
        self._overwritte  = False
        self._clean_files = True

    '''ATTRIBUTES'''
    @property
    def database(self): return self._database

    @property
    def selfHit(self):        return self._selfHit
    @selfHit.setter
    def selfHit(self, vaule): self._selfHit = value

    @property
    def hitIDformat(self): return self._hitIDformat
    @hitIDformat.setter
    def hitIDformat(self, value):
        if value not in set(['single','double','all']):
            self._hitIDformat = 'single'
        else:
            self._hitIDformat = value

    @property
    def overwritte(self):       return self._overwritte
    @overwritte.setter
    def overwritte(self, value): self._overwritte = value

    @property
    def clean_files(self):        return self._clean_files
    @clean_files.setter
    def clean_files(self, value): self._clean_files = value

    '''FUNCTIONS'''
    def add_attribute(self, attribute_value, attribute_id):
        if attribute_id in set(['-db','-outfmt']):
            raise AttributeError('The parameters in {0} cannot be altered'.format(set(['-db','-outfmt'])))

        self._exe.add_attribute(str(attribute_value), attribute_id)

    def execute_query_seq(self, sequenceID = None, sequence = None, blast_input_file = None, blast_output_file = None):
        if sequenceID is None and sequence is None:
            raise AttributeError('Either a sequence or ID is needed to perform the blast')

        if isinstance(sequenceID, (list,set,tuple)):
            raise AttributeError('Blasts can only be executed one at a time due to XML output restrictions')

        if sequenceID is None:
            sequenceID = 'QuerySequence'
        #Given only a code implies that the protein of interest is in the database itself
        if sequence is None:
            grabbedSequence = self._database.retrieve(sequenceID)
            sequenceID = grabbedSequence.id
            sequence   = grabbedSequence.sequence

        if len(re.sub(r'[Xx]','',sequence)) == 0: #All the sequence is unknown, it will crash blast
            return BR.BlastResult(queryname=sequenceID, querylength=len(sequence))

        file_prefixes    = ".".join([str(os.getpid()), str(int(time.process_time()*100000))])
        if blast_input_file is None:
            temp_input_name  = os.path.join(os.getcwd(), file_prefixes + ".tmp.fa")
        else:
            temp_input_name  = blast_input_file
        if blast_output_file is None:
            temp_output_name = os.path.join(os.getcwd(), file_prefixes + ".blast.xml.out")
        else:
            temp_output_name = blast_output_file

        QueryFasta = Fasta.build(file_name = temp_input_name, sequenceID = sequenceID, sequence = sequence, force = True)

        self._execute(input_file = QueryFasta, output_file = temp_output_name)

        BlastResult = self._parse_blast(sequence, temp_output_name)

        if self.clean_files:
            self._clean([temp_input_name,temp_output_name])

        return BlastResult

    def execute_query(self, query_file = None, blast_output_file = None):
        if isinstance(query_file, str) or isinstance(query_file, File):
            newFasta = Fasta(fasta_file = query_file)
        elif isinstance(query_file, Fasta):
            newFasta = query_file

        if newFasta.is_multifasta:
            raise BE(code = -4, value = newFasta.file.full)

        newFasta.load()
        if len(re.sub(r'[Xx]','',newFasta.sequences[0].sequence)) == 0: #All the sequence is unknown, it will crash blast
            return BR.BlastResult(queryname=newFasta.sequences[0].id, querylength=len(newFasta.sequences[0].sequence))

        if blast_output_file is None:
            temp_output_name = os.path.join(os.getcwd(), newFasta.file.prefix + "." + str(os.getpid()) + ".blast.xml.out")
        else:
            temp_output_name = blast_output_file

        self._execute(input_file = newFasta, output_file = temp_output_name)

        BlastResult = self._parse_blast(newFasta.sequences[0].sequence, temp_output_name)

        if self.clean_files:
            self._clean([temp_output_name])

        return BlastResult

    '''PRIVATE FUNCTIONS'''
    def _execute(self, input_file, output_file):
        if not os.path.isfile(output_file) or self.overwritte:
            final_executable = copy.deepcopy(self._exe)
            final_executable.add_attribute(input_file.file.full, '-query')
            final_executable.add_attribute(output_file, '-out')

            try:
                final_executable.execute()
            except SystemError as e:
                psiblast_default_warning = 'Warning: Composition-based score adjustment conditioned on sequence properties and unconditional composition-based score adjustment is not supported with PSSMs, resetting to default value of standard composition-based statistics'
                selenocysteine_warning   = 'Selenocysteine \(U\) at position'
                if not bool(re.search(psiblast_default_warning,str(e))) and not bool(re.search(selenocysteine_warning,str(e))):
                    raise BE(code = -1, value = str(e))

    def _clean(self, files):
        if not SBIglobals.debug:
            for temp_file in files:
                os.unlink(temp_file)

    def _parse_blast(self, query_sequence, blast_output_file):
        from .blast_parser import parse_blast
        return parse_blast(query_sequence, blast_output_file, self._selfHit, self._hitIDformat)

    def _check_database(self, database):
        #Database file does not exist
        if not os.path.isfile(database):
            raise BE(code = -5, value = database)

        #Database is not formated (if dbformatexe is added in the configuration path it will be autoformated)
        formatdb_files = []
        if self._search_type == 'nucl':     formatdb_sufix = ['.nhr','.nin','.nsq']
        elif self._search_type == 'prot':   formatdb_sufix = ['.phr','.pin','.psq']
        for sufix in formatdb_sufix:
            if not os.path.isfile(database + sufix):
                formatdb_files.append(database + sufix)

        if len(formatdb_files) > 0:
            try:
                dbexe  = Executable(executable    = self._configurator.get('blast','dbformatexe'),
                                    path          = self._configurator.get('blast','path'),
                                    variable_path = self._configurator.get('blast','variable_path'))

                SBIglobals.alert('debug', self, 'Trying to format de DB {0} to perform a blast search.\n'.format(database))

                dbexe.add_attribute(database, '-in')
                dbexe.add_attribute(self._search_type, '-dbtype')

                SBIglobals.alert('debug', self, 'Executing command {0}\n'.format(dbexe))

                dbexe.execute()

            except configparser.NoOptionError as e:
                raise BE(code = -6, value = formatdb_files)
            except SystemError as e:
                raise BE(code = -11, value = e)

        return Fasta(fasta_file = database)





from . import BlastResult as BR
from . import BlastHit    as BH
from SBILib.error          import BlastError  as BE
from SBILib                    import SBIglobals


from xml.dom.minidom import parseString
import sys

def parse_blast(query_sequence, blast_output_file, selfHit, hitIDformat):
    '''
        2. We retrieve the file as a Dom Object
    '''
    file_fd = open(blast_output_file,'r')
    domObj  = parseString(file_fd.read())
    file_fd.close()

    '''
        - XML parsing control variables
    '''
    error_bool = False
    error_str  = ""

    SBIglobals.alert('debug', 'blast_parser', "DEBUG: File %s opened and read correctly\n" %blast_output_file)

    '''
        3. PARSING
        3.1. Parse query identifyer
    '''
    queryname = domObj.getElementsByTagName("BlastOutput_query-def")[0].childNodes[0].nodeValue.strip().split()[0].strip()

    '''
        3.2. Parse general blast data
    '''
    querylength  = domObj.getElementsByTagName("BlastOutput_query-len")[0].childNodes[0].nodeValue
    blastversion = domObj.getElementsByTagName("BlastOutput_version")[0].childNodes[0].nodeValue
    matrix       = domObj.getElementsByTagName("Parameters_matrix")[0].childNodes[0].nodeValue
    gap_open     = domObj.getElementsByTagName("Parameters_gap-open")[0].childNodes[0].nodeValue
    gap_extend   = domObj.getElementsByTagName("Parameters_gap-extend")[0].childNodes[0].nodeValue
    db           = domObj.getElementsByTagName("BlastOutput_db")[0].childNodes[0].nodeValue

    SBIglobals.alert('debug', 'blast_parser', "DEBUG: Query is %s with length %s\n" %(queryname, querylength))

    '''
        3.3. Create the BlastResult object
    '''
    BlastOutput = BR.BlastResult(queryname     = queryname,     querylength     = querylength,
                                 blastversion  = blastversion,  blastmatrix     = matrix, 
                                 gap_open      = int(gap_open), gap_extend      = int(gap_extend),  
                                 blastdb       = db,            queryseq        = query_sequence)

    SBIglobals.alert('debug', 'blast_parser', BlastOutput.str_blast_details() + "\n")

    '''
        3.4. We start to work on each blast iteration
    '''
    for iteration in domObj.getElementsByTagName("Iteration"):

        iteration_number = int(iteration.getElementsByTagName("Iteration_iter-num")[0].childNodes[0].nodeValue)

        SBIglobals.alert('debug', 'blast_parser', "Parsing iteration %d\n" %iteration_number)

        '''
            3.4.1. And for each iteration we explore each Hit.
        '''
        for hit in iteration.getElementsByTagName("Hit"):
            '''
                 3.4.1.1. Parse de Hit identifyer and length
            '''
            hitname = hit.getElementsByTagName("Hit_def")[0].childNodes[0].nodeValue.strip()
            if hitIDformat   == 'single': hitname = hitname.split()[0].strip()
            elif hitIDformat == 'double':   hitname = " ".join(hitname.split()[:2]).strip()

            hitlength = int(hit.getElementsByTagName("Hit_len")[0].childNodes[0].nodeValue)

            SBIglobals.alert('debug', 'blast_parser', "\tParsing hit sequence %s of length %d\n" %(hitname, hitlength))

            if not same_query_hit_names(query = queryname, hit = hitname, selfHit = selfHit):
                '''
                    3.4.1.1.A. A BlastHit object is required for each possible subhit
                               Each Hit is parsed through an interanal funtion to ease read.
                '''
                for subhitlist in hit.getElementsByTagName("Hit_hsps"):

                    SBIglobals.alert('debug', 'blast_parser', "\t\tParsing subhit alignment...\n")

                    for subhit in subhitlist.getElementsByTagName("Hsp"):
                        data = parse_subhit(subhit = subhit)
                        OutputHit = BH.BlastHit(name      = hitname,    length       = hitlength,  iteration  = iteration_number,
                                                e_value   = data['ev'], align_length = data['al'], identities = data['hi'], 
                                                positives = data['hp'], gaps         = data['hg'], qseq       = data['qs'],
                                                hseq      = data['hs'], qpos         = data['qpi'],
                                                hpos      = data['hpi'],           score_seq = data['scs'])
                        '''
                             - Then each new BlastHit is added to the BlastResult
                        '''
                        BlastOutput._hits.append(OutputHit)

                        SBIglobals.alert('debug', 'blast_parser', "\t\tHit added to BlastResult List\n")

                        if not OutputHit.are_segments_ok:
                            '''
                                - We will accomulate all the errors before returning them
                            '''
                            error_bool = True
                            error_str += "WARNING: Some error has occurred on the fragmentation of the alignment "\
                            "for the query %s with %s\nERROR: PLEASE CHECK THE ALIGNMENT:\nERROR: " %(queryname, hitname)
                            error_str += "%s\n" %OutputHit

                        SBIglobals.alert('debug', 'blast_parser', "\t\tNext Sub-hit\n")

                SBIglobals.alert('debug', 'blast_parser', "\tNext Hit\n")
            else:
                SBIglobals.alert('debug', 'blast_parser', "\tHit %s is skipped as it corresponds to the query sequence\n" %hitname.split()[0].strip())
                   
    BlastOutput.set_last_iteration()

    '''
        Raise the parsing errors if any
    '''
    if error_bool:
        raise BE(code = 1, value = error_str)

    '''
        correctHitCount
    '''

    #BlastOutput.correctHitCount()
    '''
        Return BlastResult
    '''
    return BlastOutput

def parse_subhit(subhit):
    """
    Returns the required data from the given subhit 
    """
    data = {}

    """
        Totally random, Sometimes, with 0 gaps, 0 identities or 0 positives no tag is specified...
    """
    if len(subhit.getElementsByTagName("Hsp_gaps")) != 0:
        data['hg']   = int(subhit.getElementsByTagName("Hsp_gaps")[0].childNodes[0].nodeValue)
    else: data['hg'] = 0
    if len(subhit.getElementsByTagName("Hsp_identity")) != 0:
        data['hi']   = int(subhit.getElementsByTagName("Hsp_identity")[0].childNodes[0].nodeValue)
    else: data['hi'] = 0
    if len(subhit.getElementsByTagName("Hsp_positive")) != 0:
        data['hp']   = int(subhit.getElementsByTagName("Hsp_positive")[0].childNodes[0].nodeValue)
    else: data['hp'] = 0

    SBIglobals.alert('debug', 'blast_parser', "\t\tGaps: %d\n\t\tIdentities: %d\n\t\tSimilarities: %d\n" %(data['hg'], data['hi'], data['hp']))

    """
        The rest of the data is easier
    """
    data['ev']  = float(subhit.getElementsByTagName("Hsp_evalue")[0].childNodes[0].nodeValue)
    data['al']  = int(subhit.getElementsByTagName("Hsp_align-len")[0].childNodes[0].nodeValue)
    data['qs']  = str(subhit.getElementsByTagName("Hsp_qseq")[0].childNodes[0].nodeValue).strip()
    data['hs']  = str(subhit.getElementsByTagName("Hsp_hseq")[0].childNodes[0].nodeValue).strip()
    data['qpi'] = int(subhit.getElementsByTagName("Hsp_query-from")[0].childNodes[0].nodeValue)
    data['qpe'] = int(subhit.getElementsByTagName("Hsp_query-to")[0].childNodes[0].nodeValue)
    data['hpi'] = int(subhit.getElementsByTagName("Hsp_hit-from")[0].childNodes[0].nodeValue)
    data['hpe'] = int(subhit.getElementsByTagName("Hsp_hit-to")[0].childNodes[0].nodeValue)
    data['scs'] = str(subhit.getElementsByTagName("Hsp_midline")[0].childNodes[0].nodeValue).strip()

    return data

def same_query_hit_names(query, hit, selfHit):
    """
    Returns True if both names are the same (*based on the first word)
    Returns False if the names are different or skip is True
    """
    if selfHit: return False
    return query.split()[0].strip() == hit.split()[0].strip()



"""
ProteinChain

author: jbonet
date:   02/2013

@oliva's lab
"""

from ..atom     import AtomOfAminoAcid
from ..residue  import ResidueOfAminoAcid
from .          import Chain
from ..protein  import SShelper
from ..protein  import Sequencer
from SBILib.data   import aminoacids3to1

import numpy         as np


class ChainOfProtein(Chain):
    """
    A {ProteinChain} represents a collection of {AminoAcids}s
    """
    chaintype   = 'P'
    atomtype    = AtomOfAminoAcid
    resitype    = ResidueOfAminoAcid
    dictitype   = aminoacids3to1

    def __init__(self, pdb, chain):
        """
        @type  pdb: String
        @param pdb: PDB identifier

        @type  chain: String
        @param chain: Chain identifier
        """
        super(ChainOfProtein, self).__init__(pdb = pdb, chain = chain)
        self._ss_valid     = []
        self._archs        = []
        self._superarchs   = []
        self._torsionsCA   = '--'
        self._psiphi       = ''
        self._gaps         = []

    #
    # ATTRIBUTES
    #
    @property
    def aminoacids(self):
        """
        Returns the list of {AminoAcid}s
        @rtype: List of {AminoAcid}s
        """
        return self._structure

    @property
    def modified_aminoacids(self):
        for residue in self.aminoacids:
            if residue.mode == 'HETATM':
                yield residue

    @property
    def first_aminoacid(self):
        """
        Returns the first {AminoAcid}
        @rtype: {AminoAcid}
        """
        return self.aminoacids[0]

    @property
    def last_aminoacid(self):
        """
        Returns the last {AminoAcid}
        @rtype: {AminoAcid}
        """
        return self.aminoacids[-1]

    @property
    def torsionsCA(self):
        return self._torsionsCA

    @property
    def psiphi(self):
        data = ['', '', '']
        for i in range(0, len(self._psiphi), 3):
            data[0] += self._psiphi[i]
            data[1] += self._psiphi[i+1]
            data[2] += self._psiphi[i+2]
        return "\n".join(data)

    @property
    def secondary_structures(self):
        if not self.has_valid_secondary_structures:
            SShelper.locate_valid_secondarystructures(self)
        return self._ss_valid

    @property
    def archs(self):
        return self._archs

    @property
    def superarchs(self):
        return self._superarchs

    @property
    def protein_sequence(self):
        """
        Returns the exact sequence in the crystal
        @rtype: String
        """
        seq = ''
        for aa in self.aminoacids:
            seq += aa.single_letter
        return seq

    @property
    def gapped_protein_sequence(self):
        """
        Returns the sequence of the crystal with 'X' where
        the crystal contains gaps.
        @rtype: String
        """
        from ..protein import Sequencer
        return Sequencer._sequencer(self, 'seq')

    @property
    def full_protein_sequence(self):
        """
        Returns the sequence of the crystal with 'X' where
        the crystal contains gaps AND before the sequence
        if it does not start at 1.
        @rtype: String
        """
        if self.aminoacids[0].number <= 1:
            return self.gapped_protein_sequence

        seq = ''
        for x in range(self.aminoacids[0].number - 1):
            seq += 'x'
        return seq + self.gapped_protein_sequence

    @property
    def protein_idx(self):
        """
        Returns indexes (number + version) of the residues
        separated by x on gaps
        @rtype: String
        """
        return Sequencer._sequencer(self, 'idx')

    @property
    def gapped_protein_secondary_structure(self):
        return Sequencer._sequencer(self, 'str')

    #
    # BOOLEANS
    #
    @property
    def is_only_ca(self):
        counter = 0
        for aa in self.aminoacids:
            if aa.is_only_ca:
                counter += 1

        return counter >= len(self.aminoacids) * 0.8

    @property
    def has_dssp(self):
        """
        Returns True if dssp calculations have been assigned
        to the {AminoAcids} of the {ProteinChain}
        """
        return self.first_aminoacid._dssp is not None

    @property
    def has_valid_secondary_structures(self):
        return len(self._ss_valid) > 0

    @property
    def has_torsionsCA(self):
        return self._torsionsCA != '--'

    @property
    def has_psiphi(self):
        return self._psiphi != ''

    @property
    def has_archs(self):
        return len(self._archs) > 0

    @property
    def has_superarchs(self):
        return len(self._superarchs) > 0

    #
    # METHODS
    #
    def get_aminoacid(self, identifier):
        """
        Returns the {AminoAcid} with the given position identifier
        @rtype: {AminoAcid}
        """
        if isinstance(identifier, int):
            identifier = str(identifier + ' ')
        for aa in self.aminoacids:
            if aa.identifier == identifier:
                return aa

    def revert_heteroAa(self):
        for residue in self.modified_aminoacids:
            residue.normalize()

    def get_backbone_coordinates(self):
        all_coord = np.array(np.zeros(3))
        for residue in self.aminoacids:
            for atom in residue.backbone:
                all_coord = np.vstack((all_coord, atom.coordinates))
        all_coord = np.delete(all_coord, 0, 0)
        return all_coord

    def calculate_dssp(self, tmppdb = None, tmpdssp = None, cleanfiles = True):
        from ..protein import SShelper
        if not self.has_dssp:
            SShelper.calculate_dssp(self, tmppdb, tmpdssp, cleanfiles)

    def calculate_archs(self, limit_internal_ss=100,
                        limit_distance=False, allowed_gaps=0):

        if not self.has_dssp:
            self.calculate_dssp()
        if not self.has_valid_secondary_structures:
            self.secondary_structures
        if not self.has_torsionsCA:
            self.calculate_torsionsCA()
        if not self.has_psiphi:
            self.calculate_psiphi()

        SShelper.calculate_archs(self, limit_internal_ss,
                                 limit_distance, allowed_gaps)

    def calculate_torsionsCA(self, accuracy = 0):
        from ..protein import SShelper
        SShelper.calculate_torsionsCA(self, accuracy)

    def calculate_psiphi(self):
        from ..protein import SShelper
        SShelper.calculate_psiphi(self)

    #
    # REPRESENTATIONS
    #
    def archs2json(self, as_string = True):
        data = {'ID'    : self.globalID,
                'SEQ'   : self.gapped_protein_sequence,
                'STR'   : self.gapped_protein_secondary_structure,
                'ARCHS' : [x.json_format(False) for x in self.archs],
                'SUPER' : [x.json_format(False) for x in self.superarchs]}
        return repr(data) if as_string else data




"""Path

author: jbonet
date:   02/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
import os
import fnmatch
import shutil

"""
Dependences in SBI library
"""
from SBILib.error import FileError
from .             import File
from SBILib           import SBIglobals

class Path(object):
    """Path

        Path is a collection of @staticmethod (can be called without declaring an instance of the class) to 
        help in file and directory management.

        Methods:
            > list_files()      : Returns any file in a directory tree matching a specific pattern.
                                    - root              (string): Root of the directory tree to search
                                                                  @Default: current working directory
                                    - pattern           (string): Expression to match (ls-like format)
                                                                  (Accepts list of strings)
                                                                  @Default: *
                                    - avoid_empty_files (bool)  : Ignore files with size 0
                                                                  @Default: True
                                    - rootless          (bool)  : When False, the name of the files are returned with
                                                                  absolute path. Otherwise the root is removed.
                                                                  @Default: False
                                  @Yields file names

            > list_directories(): Returns all dirs in a directory tree.
                                    - root              (string): Root of the directory tree to search
                                                                  @Default: current working directory
                                    - rootless          (bool)  : When False, the name of the dirs are returned with
                                                                  absolute path. Otherwise the root is removed.
                                                                  @Default: False
                                  @Yields directory names

            > sync_directories(): Syncronizes two directory trees.
                                    - sourcedir      (string): Name of the original directory
                                                               @Mandatory
                                    - destinationdir (string): Name of the syncronized directory
                                                               @Mandatory
                                    - by_dir         (bool)  : Sync is performed at directory level
                                                               @Default: True
                                    - by_file        (bool)  : Sync is performed at file level
                                                               @Default: False
                                    - listdif        (bool)  : When True, sync is not performed, only listed
                                                               @Default: False
                                    @Yields file/dir names when listdif is True
                                    @Raises AttributeError if by_dir and by_file are both True

            > do_files_match()  : Determines if there is any file in a directory tree matching a specific pattern.
                                    - root              (string): Root of the directory tree to search
                                                                  @Default: current working directory
                                    - pattern           (string): Expression to match (ls-like format)
                                                                  @Default: *
                                    - avoid_empty_files (bool)  : Ignore files with size 0
                                                                  @Default: True
                                  @Returns Boolean

            - mkdir()           : Creates a new directory. 
                                  Ignores the command if the directory already exists.
                                  Creates parent directories as required.
                                  Recipe: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/82465
                                    - newdir (string): Name of the new directory.
                                                       @Mandatory
                                  @Raises FileError

            - copy_file()       : Copy a file from one place to another.
                                    - source      (string): Name of the original file
                                                            @Mandatory
                                    - destination (string): Name of the new copied file
                                                            @Mandatory
                                    -overwrite    (bool)  : Ignores if a file with the same name existed before
                                                            @Default: False
                                  @Raises FileError
    """
    @staticmethod
    def list_files(root = os.curdir, pattern = '*', avoid_empty_files = True, rootless = False):
        """
            > list_files()      : Returns any file in a directory tree matching a specific pattern
                                    - root              (string): Root of the directory tree to search
                                                                  @Default: current working directory
                                    - pattern           (string): Expression to match (ls-like format)
                                                                  (Accepts list of strings)
                                                                  @Default: *
                                    - avoid_empty_files (bool)  : Ignore files with size 0
                                                                  @Default: True
                                    - rootless          (bool)  : When False, the name of the files are returned with
                                                                  absolute path. Otherwise the root is removed.
                                                                  @Default: False
                                  @Yields file names
        """
        if os.path.isfile(root): yield root

        search_patterns = []
        if not isinstance(pattern, list):
            search_patterns.append(pattern)
        else:
            search_patterns = pattern
        for pat in search_patterns:
            for path, dirs, files in os.walk(os.path.abspath(root)):
                for filename in fnmatch.filter(files, pat):
                    if not avoid_empty_files or os.path.getsize(os.path.join(path, filename)) > 0:
                        SBIglobals.alert('debug', Path(), 'Found file {0}'.format(os.path.join(path,filename)))
                        if not rootless:
                            yield os.path.join(path, filename)
                        else:
                            root = os.path.abspath(root) + "/"
                            yield os.path.join(path, filename).replace(root,'')

    @staticmethod
    def list_directories(root = os.curdir, rootless = False):
        """
            > list_directories(): Returns all dirs in a directory tree
                                    - root              (string): Root of the directory tree to search
                                                                  @Default: current working directory
                                    - rootless          (bool)  : When False, the name of the dirs are returned with
                                                                  absolute path. Otherwise the root is removed.
                                                                  @Default: False
                                  @Yields directory names
        """
        for path, dirs, files in os.walk(os.path.abspath(root)):
            for onedir in dirs:
                SBIglobals.alert('debug', Path(), 'Found directory {0}'.format(os.path.join(path,onedir)))
                if not rootless:
                    yield os.path.join(path,onedir)
                else:
                    yield os.path.join(path,onedir).replace(root, '')

    @staticmethod
    def sync_directories(sourcedir, destinationdir, by_dir = True, by_file = False, listdif = False):
        """
            > sync_directories(): Syncronizes two directory trees.
                                    - sourcedir      (string): Name of the original directory
                                                               @Mandatory
                                    - destinationdir (string): Name of the syncronized directory
                                                               @Mandatory
                                    - by_dir         (bool)  : Sync is performed at directory level
                                                               @Default: True
                                    - by_file        (bool)  : Sync is performed at file level
                                                               @Default: False
                                    - listdif        (bool)  : When True, sync is not performed, only listed
                                                               @Default: False
                                    @Yields file/dir names when listdif is True
                                    @Raises AttributeError if by_dir and by_file are both True
        """
        if by_file is True and by_dir is True:
            raise AttributeError('Both sync methods can not be active simultaneously')

        source_dirs = set()
        if by_dir:
            for onedir in Path.list_directories(root = sourcedir, rootless = True):
                source_dirs.add(onedir)
            for onedir in Path.list_directories(root = destinationdir, rootless = True):
                if onedir in source_dirs:
                    source_dirs.remove(onedir)
        if by_file:
            for onedir in Path.list_files(root = sourcedir, rootless = True):
                source_dirs.add(onedir)
            for onedir in Path.list_files(root = destinationdir, rootless = True):
                if onedir in source_dirs:
                    source_dirs.remove(onedir)

        previous_dir   = '#not_a_dir_at_all'
        sourcedir      = os.path.abspath(sourcedir)
        destinationdir = os.path.abspath(destinationdir)
        for onedir in sorted(source_dirs):
            onedir = onedir.lstrip('/')
            if not os.path.commonprefix([previous_dir, onedir]) == previous_dir:
                SBIglobals.alert('verbose', Path(), '{0} is different from {1} to {2}'.format(onedir,sourcedir,destinationdir))
                if not listdif:
                    fullori = os.path.join(sourcedir,onedir)
                    shutil.copytree(os.path.join(sourcedir,onedir),os.path.join(destinationdir,onedir))
                else:
                    yield os.path.join(sourcedir, onedir)
                previous_dir = onedir

    @staticmethod
    def do_files_match(root = os.curdir, pattern = '*', avoid_empty_files = True):
        """
            > do_files_match()  : Determines if there is any file in a directory tree matching a specific pattern
                                    - root              (string): Root of the directory tree to search
                                                                  @Default: current working directory
                                    - pattern           (string): Expression to match (ls-like format)
                                                                  @Default: *
                                    - avoid_empty_files (bool)  : Ignore files with size 0
                                                                  @Default: True
                                  @Returns Boolean
        """
        if os.path.isfile(root): return True

        search_patterns = []
        if not isinstance(pattern, list):
            search_patterns.append(pattern)
        else:
            search_patterns = pattern
        for pat in search_patterns:
            for path, dirs, files in os.walk(os.path.abspath(root)):
                for filename in fnmatch.filter(files, pat):
                    if not avoid_empty_files or os.path.getsize(os.path.join(path, filename)) > 0:
                        return True
        return False

    @staticmethod
    def mkdir(newdir):
        """
            - mkdir()           : Creates a new directory. 
                                  Ignores the command if the directory already exists.
                                  Creates parent directories as required.
                                  Recipe: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/82465
                                    - newdir (string): Name of the new directory.
                                                       @Mandatory
                                  @Raises FileError
        """
        if os.path.isdir(newdir): 
            SBIglobals.alert('debug', Path(), 'A directory with that name already exists.')
            pass
        elif os.path.isfile(newdir):
            raise FileError(3,newdir,'filewithname');
        else:
            SBIglobals.alert('debug', Path(), 'Creating dir {0}.'.format(newdir))
            head, tail = os.path.split(newdir)
            if head and not os.path.isdir(head):
                Path.mkdir(head)
            if tail:
                os.mkdir(newdir)

    @staticmethod
    def copy_file(source, destination, overwrite = False):
        """
            - copy_file()       : Copy a file from one place to another.
                                    - source      (string): Name of the original file
                                                            @Mandatory
                                    - destination (string): Name of the new copied file
                                                            @Mandatory
                                    -overwrite    (bool)  : Ignores if a file with the same name existed before
                                                            @Default: False
                                  @Raises FileError
        """
        # Local overwrite takes precedence over Global overwrite
        overwrite = SBIglobals.decide_overwrite(overwrite)

        if not os.path.exists(source):
            raise FileError(3, source, 'noexists')
        if os.path.exists(destination) and not overwrite:
            raise FileError(3, destination, 'exists')

        shutil.copyfile(src = source, dst = destination)



"""Enzymelink

author: jbonet
date:   10/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
import os, re
import urllib.request, urllib.parse, urllib.error

"""
Dependences in SBI library
"""
from SBILib.databases import Enzymeftp
from SBILib.beans.Path     import Path
from SBILib.beans.File     import File

class Enzymelink(object):
    """The Enzymelink class controls the download and parsing of Enzyme database

    """
    def __init__(self, local = None):
        self._local   = os.path.abspath(local)
        self._dfile   = 'enzyme.dat'
        self._cfile   = 'enzclass.txt'
        self.__name__ = 'databases.Enzymelink'    # This must be included in every class for the SBIglobals.alert()
        self._enzfile = 'enzyme.gz'
        if local is not None:
            self.local = local

    """ATTRIBUTES"""
    @property
    def local(self):         return self._local
    @local.setter
    def local(self, value):
        self._local   = os.path.abspath(value)
        self._dfile   = os.path.join(self._local, self._dfile)
        self._cfile   = os.path.join(self._local, self._cfile)
        self._enzfile = os.path.join(self._local, self._enzfile)

    @property
    def localEnzymes(self):
        enzFile = File(self._enzfile, 'r')
        for enz_line in enzFile.descriptor:
            yield enz_line

    @property
    def source(self):
        return Enzymeftp['show']

    """BOOLEANS"""
    @property
    def has_local(self):    return self._local is not None

    """METHODS"""
    def download(self):
        if not self.has_local:
            raise NameError('A local Enzyme database directory must be defined.')

        Path.mkdir(self.local)
        urllib.request.urlretrieve(Enzymeftp['dat'], self._dfile)
        urllib.request.urlretrieve(Enzymeftp['cls'], self._cfile)

        self._process()

        return True

    def get_Enzyme(self, enzyme):
        if self.has_local:
            for enz_line in self.localEnzymes:
                if enz_line.split('\t')[0] == enzyme:
                    return enz_line
        else:
            raise NameError('A local Enzyme database directory must be defined.')

    def get_Enzymes(self, ENZset):
        if isintance(ENZset, str):
            warnings.warn('For single enzyme search the get_Enzyme function is recomended.')
            yield self.get_Enzyme(ENZset)

        if self.has_local:
            for enz_line in self.localEnzymes:
                if enz_line.split('\t')[0] in ENZset:
                    yield enz_line
        else:
            raise NameError('A local Enzyme database directory must be defined.')

    """PRIVATE METHODS"""
    def _process(self):
        enzymes = self._parse_enzclass() + self._parse_enzymedat()
        enzymes.sort()

        enzFile = File(self._enzfile,'w', True)
        for e in enzymes:
            enzFile.write(repr(e)+"\n")
        enzFile.close()

    def _parse_enzclass(self):
        enzymes = []
        fd = open(self._cfile, 'r')
        for line in fd:
            if re.match(r'^\d', line):
                line = re.sub(r'\. ','.', line)
                line = re.sub('\'', '\\\'', line)
                data = line.split()
                enzymes.append(Enzyme(ec = data[0], description = ' '.join(data[1:])))
        fd.close()
        return enzymes

    def _parse_enzymedat(self):
        code          = None
        name          = None
        cofr          = ''
        reac          = ''
        reactions_dic = {}
        enzymes       = []

        cofactor_separator = re.compile(r"; | or |;| and ")
        multiple_reac_sepr = re.compile(r"\.\(\d+\)\s")
        noisy_words_equal  = re.compile(r"^A |^An |secondary | a | an ")
        unspaced_summator  = re.compile(r" \+|\+ ")
        unspaced_equalate  = re.compile(r" \=|\= ")
        summator_separator = re.compile(r" \+ | or ")
        equalate_separator = re.compile(r" \= ")
        cardinal_numerical = re.compile(r"^(\d+)\s+")

        fd = open(self._dfile)
        for line in fd:
            line = re.sub('\'', '\\\'', line)
            line = re.sub('<element>','', line)
            line = re.sub('</element>','', line)
            if line.startswith('//'):
                if code is not None:
                    enzymes.append(Enzyme(code, name))
                    if cofr != '':
                        cofr = cofr[:-1]
                        for fct in cofactor_separator.split(cofr):
                            enzymes[-1].cofactors = fct

                    if reac != '':
                        if re.search(r"\=", reac):
                            for rct in multiple_reac_sepr.split(reac):
                                repeated = False
                                if rct.startswith('(1)'): rct = rct[4:]
                                if rct.endswith('.'):     rct = rct[:-1]
                                rct = re.sub(noisy_words_equal, '', rct)
                                rct = re.sub(unspaced_summator, ' + ', rct)
                                rct = re.sub(unspaced_equalate, ' = ', rct)
                                rct = re.sub('\s+',' ', rct)
                                rct = re.sub('\> \=', '\>\=', rct)
                                rct = re.sub('\< \=', '\<\=', rct)
                                rct = re.sub('\(n \= ', '(n=', rct)

                                if not rct in reactions_dic:
                                    rid = len(reactions_dic) + 1
                                    reactions_dic[rct] = rid
                                else:
                                    rid = reactions_dic[rct]
                                    repeated = True

                                enzymes[-1].reactions.append(Reaction(description = rct, repeated = repeated))
                                if repeated: continue
                                try:
                                    (substrate, product) = equalate_separator.split(rct)
                                except:
                                    (substrate, product) = unspaced_equalate.split(rct)
                                cardinality = 1
                                for subs in summator_separator.split(substrate):
                                    cardinal = cardinal_numerical.search(subs)
                                    if cardinal:
                                        cardinality = cardinal.group(1).strip()
                                        subs = re.sub(cardinal_numerical, '', subs)
                                        subs.strip()
                                    enzymes[-1].reactions[-1].substrates.append((subs,cardinality))
                                    cardinality = 1
                                for prod in summator_separator.split(product):
                                    cardinal = cardinal_numerical.search(prod)
                                    if cardinal:
                                        cardinality = cardinal.group(1).strip()
                                        prod = re.sub(cardinal_numerical, '', prod)
                                        prod.strip()
                                    enzymes[-1].reactions[-1].products.append((prod,cardinality))
                                    cardinality = 1

                        else:
                            if not reac in reactions_dic:
                                rid = len(reactions_dic) + 1
                                reactions_dic[reac] = rid
                            else:
                                rid = reactions_dic[reac]
                            enzymes[-1].reactions.append(Reaction(description = reac))
                    if len(unip) > 0:
                        enzymes[-1].uniprots = unip

                code = None
                name = None
                cofr = ''
                reac = ''
                unip = []
                continue
            if line.startswith('ID'):
                code = line.split()[-1].strip()
                name = ""
                continue
            if line.startswith('DE'):
                if name != "":
                    name += " "
                name += " ".join(line.split()[1:]).strip()
                # name = name[:-1]
                continue
            if line.startswith('CF'):
                cofactors = " ".join(line.split()[1:]).strip()
                cofr = cofr + cofactors
                continue
            if line.startswith('CA'):
                reactions = " ".join(line.split()[1:]).strip()
                reac = reac + reactions
                continue
            if line.startswith('DR'):
                unipcodes = [x.replace(' ','').split(',')[1] for x in line[2:].strip().split(';')[:-1]]
                unip.extend(unipcodes)
                continue
        fd.close()
        return enzymes

class Enzyme(object):

    def __init__(self, ec = None, description = None, inline = None):
        if inline is not None:
            inline = inline.strip().split('\t')
        self.ec          = ec.replace(' ','')     if inline is None else inline[0]
        self.description = description.strip()    if inline is None else inline[1]
        self.deleted     = self.description == 'Deleted entry.'
        self.level       = 4 - self.ec.count('-') if inline is None else int(inline[2])
        self.parents     = []                     if inline is None else eval(inline[3])
        self.direct_p    = 'NULL'                 if inline is None else inline[4]
        self._cofactors  = []                     if inline is None else eval(inline[5])
        self.transfers   = []                     if inline is None else eval(inline[6])
        self.uniprots    = []                     if inline is None else eval(inline[7])
        self.reactions   = []

        if inline is None:
            self._parents()
            self._transfers()
        else:
            self._reactions(eval(inline[8]))

    @property
    def cofactors(self):
        return self._cofactors
    @cofactors.setter
    def cofactors(self, value):
        if re.search('\(',value) and not re.search('\)',value): value = re.sub('\(','',value)
        if re.search('\)',value) and not re.search('\(',value): value = re.sub('\)','',value)
        self._cofactors.append(value.upper())

    @property
    def compounds(self):
        for reaction in self.reactions:
            for subs in reaction.substrates:
                yield subs[0]
            for prod in reaction.products:
                yield prod[0]

    @property
    def has_transfers(self):
        return len(self.transfers) != 0
    @property
    def has_direct_parent(self):
        return self.direct_p != 'NULL'
    @property
    def is_deleted(self):
        return self.deleted

    def _reactions(self, reactarray):
        for react in reactarray:
            self.reactions.append(Reaction(inline = react))

    def _parents(self):
        self.parents.append(self.ec)
        pieces = self.ec.split('.')
        for i in range(3,0,-1):
            if pieces[i] != '-':
                pieces[i] = '-'
                parent = '.'.join(pieces)
                self.parents.append(parent)
                pieces = parent.split('.')
        try:
            self.direct_p = self.parents[1]
        except:
            self.direct_p = 'NULL'
        self.parents.reverse()

    def _transfers(self):
        if self.description.startswith('Transferred entry'):
            d = self.description
            d.strip('.')
            aggregator_separator = re.compile(r",|and")
            self.transfers = [x.strip().rstrip('.') for x in aggregator_separator.split(d.split(':')[1])]

    def __lt__(self, other):
        me = [int(x) for x in self.ec.replace('-','0').replace('n','1000').split('.')]
        he = [int(x) for x in other.ec.replace('-','0').replace('n','1000').split('.')]

        for i in range(4):
            if me[i] > he[i]:
                return False
            elif me[i] < he[i]:
                return True
        return False

    def __gt__(self, other):
        me = [int(x) for x in self.ec.replace('-','0').replace('n','1000').split('.')]
        he = [int(x) for x in other.ec.replace('-','0').replace('n','1000').split('.')]

        for i in range(4):
            if me[i] > he[i]:
                return True
            elif me[i] < he[i]:
                return False
        return False

    def __repr__(self):
        data = []
        data.append('{0.ec}\t{0.description}\t{0.level}\t{0.parents}\t{0.direct_p}'.format(self))
        data.append('{0.cofactors}\t{0.transfers}\t{0.uniprots}'.format(self))
        data.append('{0.reactions}'.format(self))

        return "\t".join(data)
        # return '{0.ec}\t{0.description}\t{0.level}\t'.format(self) + repr(self.parents) + repr(self.reactions)

class Reaction(object):

    def __init__(self, description = None, repeated = False, inline = None):
        self.description = description.rstrip('\\') if inline is None else inline[0]
        self.substrates  = []                       if inline is None else inline[1]
        self.products    = []                       if inline is None else inline[2]
        self.repeated    = repeated                 if inline is None else inline[3]

    def __repr__(self):
        data = []
        data.append(self.description)
        data.append(self.substrates)
        data.append(self.products)
        data.append(self.repeated)
        return '{0}'.format(data)
        # return '{0.description},{0.substrates},{0.products},{0.repeated}'.format(self)




# -*- coding: utf-8 -*-
"""
.. codeauthor:: Jaume Bonet <jaume.bonet@gmail.com>

.. affiliation::
    Structural BioInformatics Lab <sbi.upf.edu>
    Baldo Oliva <baldo.oliva@upf.edu>

.. class:: ChainFrame
.. class:: ProteinChainFrame
.. class:: NucleotideChainFrame
"""

# Standard Libraries
import warnings
import os
import re
from typing import Optional, Union, TypeVar
from tempfile import gettempdir
from itertools import groupby

# External Libraries
import pandas as pd
import numpy as np

# This Library
from ..Frame3D import Frame3D
from SBILib.core import core
from SBILib.data import alphabet

__all__ = ['ChainFrame', 'ProteinChainFrame', 'NucleotideChainFrame']

CF = TypeVar('CF', bound='ChainFrame')


class ChainFrame( Frame3D ):
    """A DataFrame container of single chain data. A protein fragment
    would also fall into this category.
    """
    _subtyp = 'chain_frame'

    def __init__( self, *args, **kw ):
        super(ChainFrame, self).__init__(*args, **kw)

    # Attributes
    @property
    def globalID( self ):
        """
        Full identifier of the Chain.
        By default, it would be ``<pdb-name>_<chain-id>``

        :returns: :class:`str`
        """
        warnings.warn(".globalID will be deprecated. use .id instead", FutureWarning, stacklevel=2)
        return self.id

    @property
    def id( self ):
        """Structure identifier.

        :returns: :class:`str`
        """
        return "{0}_{1}".format(self.pdb, self.chain)

    @property
    def pdb( self ):
        """
        PDB identifier.

        :returns: :class:`str`
        """
        return self._id

    @property
    def chain( self ):
        """Chain identifier.

        :returns: :class:`str`
        """
        return str(self._current_model[self._current_asym].unique()[0])

    @property
    def residue_count( self ):
        """Number of residues in the :class:`.ChainFrame` (of the first *-main-* entity).

        .. warning::
            This substitutes :func:`len` in pre-pandas versions of the code,
            as overwritting :func:`len` is problematic.

        :return: :class:`int`
        """
        try:
            entity = self._current_model.iloc[0]['label_entity_id']
            entity = self._current_model[self._current_model['label_entity_id'] == entity]
            return len(entity.groupby(self._residue_columns, sort=False))
        except IndexError:
            # Produced for no .iloc[0], which means there is no data here.
            return 0

    @property
    def compounds( self ):
        """List the compounds in the :class:`.ChainFrame`.

        :yields: :class:`.ResidueFrame`
        """
        for _ in self.groupby(self._residue_columns + [self._current_comp, self._current_asym], sort=False):
            yield _[1]

    @property
    def first_compound( self ):
        """Obtain the first compound of the :class:`.ChainFrame`.

        :return: :class:`.ResidueFrame`
        """
        return list(self.compounds)[0]

    @property
    def last_compound( self ):
        """Obtain the first compound of the :class:`.ChainFrame`.

        :return: :class:`.ResidueFrame`
        """
        return list(self.compounds)[-1]

    # Boolean
    @property
    def is_ordered( self ):
        """Checks if the residue numbering of the chain is ordered; considering insertion codes too.

        Ignores waters and ligands.

        .. note::
            This functionality depends on the :ref:`global configuration options <configuration>` ``structure.source``.

        :return: :class:`bool`
        """
        if core.get_option('structure', 'source') == 'label':
            return True  # label NEVER is unoerdered
        else:
            if not self.has_insertion_codes:
                df = self.dehydrate(False).remove_heteroatoms(False)
                return pd.Index(df[df._current_seq].loc[df[df._current_seq].shift() != df[df._current_seq]]).is_monotonic
            return False  # Insertion codes will always mean some type of "disorder" in the sequence.

    # Methods
    def renumber( self, start=1, inplace=True ):
        """Renumber the residues of :class:'.ChainFrame'.

        .. note::
            This functionality depends on the :ref:`global configuration options <configuration>` ``structure.source``.
        """
        src = core.get_option('structure', 'source')

        df = self

        gf = [list(group) for key, group in groupby(df[df._residue_columns].values.tolist())]
        max_value = len(gf)
        atoms = np.asarray([(len(x)) for x in gf])

        numbers = [[x, ] * atoms[i] for i, x in enumerate(range(start, start + max_value))]

        df[df._current_seq] = np.asarray(numbers).flatten()
        if src == 'auth' and 'pdbx_PDB_ins_code' in df.columns:
            df['pdbx_PDB_ins_code'] = np.asarray(['', ] * df.shape[0])

        df['id'] = list(range(1, df.shape[0] + 1))

        return self._inplace(df, inplace, False)

    def B_factor( self, values, inplace=False ):
        """Substitue the B-factor values for the provided ones.

        This is useful as coloring by B-factor is allowed in multiple protein representation
        softwares and allows plotting through other data of interest.

        :param values: Values to replace the current B-factor. If a single value is provided,
            all residues are given the same B-factor. If a number of values equal to the number
            of residues is provided, each residue gets the same value for all atoms. If a number
            of values equal to the atoms in the residue are provided, each atom gets its own value.
        :type values: Union[:class:`int`, :func:`list`]

        :return: :class:`.ResidueFrame`

        :raise:
            :AttributeError: If the number of values given do not match any of the accepted
                conditions.
        """
        df = self
        if isinstance(values, (int, float)):
            df['B_iso_or_equiv'] = [values, ] * self.shape[0]
        elif isinstance(values, (list, np.ndarray)):
            if len(values) == 1:
                df['B_iso_or_equiv'] = [values[0], ] * self.shape[0]
            elif len(values) == self.shape[0]:
                df['B_iso_or_equiv'] = values
            elif len(values) == self.compounds:
                df2 = []
                for i, aa in enumerate(df.compounds):
                    df2.append(aa.B_factor(values[i], inplace=True))
                df = pd.concat(df2)
            else:
                raise AttributeError('Number of values cannot be fitted to any definition')
        return self._inplace(df, inplace)

    def pop_out( self, key: Optional[Union[int, str]] = 0 ) -> CF:
        """Delete the requested position by order from the :class:`.ChainFrame`.

        .. note::
            This functionality depends on the :ref:`global configuration option <configuration>` ``structure.model``.
            This functionality depends on the :ref:`global configuration option <configuration>` ``structure.source``.

        :param key: Array like access of the target position.

        :return: The new :class:`.ChainFrame` without the requested position.

        :raise:
            :IndexError: If the position requested cannot be found (bigger than ``count(self)``).
        """
        df = self._current_model
        cl = self._residue_columns
        dd = [x for x, _ in df.groupby(cl, sort=False)][key]
        if len(cl) == 1:
            idx = df[df[cl[0]] == dd[0]].index
        else:
            idx = df[(df[cl[0]] == dd[0]) & (df[cl[1]] == dd[1])].index
        return df.drop(idx)

    # Private Pandas Methods
    def _inheritance( self, other ):
        self._id = other._id
        return self


class ProteinChainFrame( ChainFrame ):
    _subtyp = 'protein_chain_frame'

    def __init__( self, *args, **kw ):
        super(ProteinChainFrame, self).__init__(*args, **kw)
        self._aalimit = None

    # Attributes
    @property
    def aminoacids( self ):
        """Iterate over the compounds in the :class:`.ChainFrame`.

        :yields: :class:`.ProteinResidueFrame`
        """
        df = self.remove_heteroatoms(False).dehydrate(False)
        df.set_index(df._residue_columns, drop=False, inplace=True)
        for _ in df.groupby(level=[0, 1], axis=0, sort=False):
            yield _[1]

    @property
    def first_aminoacid( self ):
        """Obtain the first amino acid.

        :return: :class:`.ProteinResidueFrame`
        """
        return self.first_compound

    @property
    def last_aminoacid(self):
        """Obtain the first amino acid.

        :return: :class:`.ProteinResidueFrame`
        """
        for residue in reversed(list(self.aminoacids)):
            return residue

    @property
    def protein_sequence(self):
        """Returns the sequence of the crystalized residues.

        It considers all the residues with density in the 3D structure and has no
        consideration for whether or not there might be gaps in the structure.

        :return: :class:`str`

        .. seealso::
            :meth:`.ProteinChainFrame.gapped_protein_sequence`
            :meth:`.ProteinChainFrame.full_protein_sequence`
        """
        df = self.remove_heteroatoms(False).dehydrate(False)
        df = df.set_index(df._residue_columns, drop=False)
        lvl = list(range(len(df._residue_columns)))
        seq = df.groupby(level=lvl, axis=0, sort=False).first()[df._current_comp].values
        return "".join([alphabet.aminoacids3to1(code3) for code3 in seq])

    @property
    def gapped_protein_sequence(self):
        """Returns the sequence of the crystalized residues filling with 'X' missed densities.

        Missing densities are calculated by evaluating for each two residues ir they follow each other or not.

        :return: :class:`str`

        .. seealso::
            :meth:`.ProteinChainFrame.protein_sequence`
            :meth:`.ProteinChainFrame.full_protein_sequence`
        """
        from ..protein import Sequencer
        return Sequencer._sequencer(self, 'seq')

    @property
    def full_protein_sequence( self ):
        """Returns the sequence of the crystalized residues filling with 'X' missed densities and starting
        non-crystalized residues.

        Takes into account the starting number of the chain and considers that all previous residues are missing.

        :return: :class:`str`

        .. seealso::
            :meth:`.ProteinChainFrame.protein_sequence`
            :meth:`.ProteinChainFrame.gapped_protein_sequence`
        """
        if self.first_aminoacid.number <= 1:
            return self.gapped_protein_sequence

        seq = ''
        for x in range(self.first_aminoacid.number - 1):
            seq += 'x'
        return seq + self.gapped_protein_sequence

    # Methods
    def calculate_dssp( self, tmppdb=None, tmpdssp=None, cleanfiles=True, minimize=False, simplify='BGITS' ):
        """Generate the [DSSP](https://github.com/cmbi/xssp) assignation for the :class:`.ProteinChainFrame`.

        .. warning::
            If the object already contains DSSP data,

        :param str tmppdb: Name for the PDB file. If :data:`None` provided, a temporary file is created.
        :param str tmpdssp: Name for the DSSP file. If :data:`None` provided, a temporary file is created.
        :param bool cleanfiles: If :data:`True`, PDB and DSSP files are deleted after execution.
        :param bool minimize: If :data:`True`, ignore data from DSSP output that is not secondary structure
            assignation or accessibility.
        :param str simplify: Secondary structure codes to simplify. ``H``, ``E`` and ``L`` are ignored if provided,
            as they cannot be simplified. The default represents the maximum possible simplification.

        :return: :class:`.DSSPFrame`
        """
        from SBILib.external import DSSPExe

        tmppdb = tmppdb if tmppdb is not None else os.path.join(gettempdir(), '{}.pdb'.format(os.getpid()))
        tmpdssp = tmpdssp if tmpdssp is not None else os.path.join(gettempdir(), '{}.dssp'.format(os.getpid()))
        if not os.path.isfile(tmppdb):
            self.dehydrate(False).remove_heteroatoms(False).write(tmppdb, format='pdb', force=False, clean=False)
        dssp = DSSPExe(pdb=tmppdb, dssp=tmpdssp, cleanpdb=cleanfiles, cleandssp=cleanfiles, minimize=minimize).dsspdata.simplify(simplify)

        return dssp.copy()

    def split_discontinuous( self ):
        """Find the discontinuous pieces of :class:`.ProteinChain` and separate them.

        :return: :func:`list` of :class:`.Frame3D`
        """
        gps = re.split(r'x+', self.gapped_protein_sequence)
        ranges = [[1, len(gps[0])]]
        base = 0
        for i in range(1, len(gps)):
            base += len(gps[i - 1])
            ranges.append([base + 1, base + len(gps[i])])
        with core.on_option_value('structure', 'source', 'label'):
            for i, r in enumerate(ranges):
                ranges[i] = self['Residue:{0}-{1}'.format(r[0], r[1])]
        return ranges


class NucleotideChainFrame( ChainFrame ):
    _subtyp = 'nucleotide_chain_frame'

    def __init__( self, *args, **kw ):
        super(NucleotideChainFrame, self).__init__(*args, **kw)



from SBILib.data import crystal_method_has_resolution
from . import process_EXPERIMENT_line, process_RESOLUTION_line
from . import process_RFACTOR_line, process_FREER_line
from SBILib.beans.JSONer import JSONer


class Experiment(JSONer):
    def __init__(self, line):
        self._xpdta      = process_EXPERIMENT_line(line)
        self._res        = None
        self._rfactor    = None
        self._freeR      = None

    #
    # ATTRIBUTES
    #
    @property
    def xpdta(self):
        return self._xpdta

    @property
    def resolution(self):
        return float(self._res) if self._res is not None else -1

    @property
    def rfactor(self):
        if self._rfactor is None:
            if self.has_resolution:
                return float(0)
            else:
                return 'NULL'
        return float(self._rfactor)

    @property
    def freeR(self):
        if self._freeR is None:
            if self.has_resolution:
                return float(0)
            else:
                return 'NULL'
        return float(self._freeR)

    #
    # BOOLEANS
    #
    @property
    def has_resolution(self):
        return self.xpdta in crystal_method_has_resolution

    #
    # READ FUNCTIONS
    #
    def update_experiment(self, line):
        self._xpdta += ' ' + process_EXPERIMENT_line(line)

    def add_resolution(self, line):
        self._res = process_RESOLUTION_line(line)

    def add_rfactor(self, line):
        if self.has_resolution and self._rfactor is None:
            self._rfactor = process_RFACTOR_line(line)

    def add_freer(self, line):
        if self.has_resolution and self._freeR is None:
            self._freeR = process_FREER_line(line)

    #
    # OUT FUNCTIONS
    #
    def as_dict(self):
        nobj = {}
        nobj['xpdta']      = self.xpdta
        nobj['resolution'] = self.resolution
        nobj['rfactor']    = self.rfactor
        nobj['freer']      = self.freeR
        return nobj

    def __repr__(self):
        return self.json()



"""
Atom

author: jbonet
date:   02/2013

@oliva's lab
"""
import numpy as np

class Atom(object):
    """
    An {Atom} is simply a point in space defined by 3 coordinates
    """

    backbone_atoms = set()
    
    def __init__(self, number, name, x, y, z, occupancy="", tempFactor="", element="", charge=""):
        """
        @type  number: Integer
        @param number: Atom number
        
        @type  name: String
        @param name: Atom identifier
        
        @type  x: Decimal
        @param x: x coordinate
        
        @type  y: Decimal
        @param y: y coordinate
        
        @type  z: Decimal
        @param z: z coordinate
        """
        self._number = int(number)
        self._name   = name
        self._coordinates = np.array([float(x), float(y), float(z)])
        self._occupancy = occupancy
        self._tempFactor = tempFactor
        self._element = element
        self._charge = charge
    #
    # ATTRIBUTES
    #
    @property
    def number(self):
        """
        Atom number
        @rtype: Integer
        """
        return int(self._number)
    
    @number.setter
    def number(self, value):
        self._number = int(value)
        
    @property
    def name(self):
        """
        Atom name
        @rtype: String
        """
        return self._name
    
    @property
    def pretty_name(self):
        """
        Returns a version of the name adapted to print in PDB
        The String has a fixed length for that matter
        @rtype: String{4}
        """
        if len(self._name) <= 3:
            return " "  + self._name.ljust(3)
        else:
            return self._name.ljust(4)
        
    @property
    def x(self):
        """
        x coordinate
        @rtype: Decimal
        """
        return self._coordinates[0]
    
    @property
    def y(self):
        """
        y coordinate
        @rtype: Decimal
        """
        return self._coordinates[1]
    
    @property
    def z(self):
        """
        z coordinate
        @rtype: Decimal
        """
        return self._coordinates[2]
    
    @property
    def coordinates(self):
        """
        @rtype: numpy.array
        """
        return self._coordinates

    @property
    def occupancy(self):
        """
        @rtype: Decimal or None
        """
        return self._occupancy

    @property
    def tempFactor(self):
        """
        @rtype: Decimal or None
        """
        return self._tempFactor

    @property
    def element(self):
        """
        @rtype: String or None
        """
        return self._element

    @property
    def charge(self):
        """
        @rtype: String or None
        """
        return self._charge  

    @property
    def is_backbone(self):
      """
      Checks if the Atom is part of the backbone of the residue
      @rtype: Boolean
      """
      return self._name in self.backbone_atoms
    
    @property
    def is_sidechain(self):
      """
      Checks if the Atom is part of the side chain of the residue
      @rtype: Boolean
      """
      return not self.is_backbone if len(self.backbone_atoms) > 0 else False
    
    #
    # METHODS
    #
    def rotate(self, matrix = None):
        """
        Rotates an atom according to a given matrix
        """
        if matrix is None: matrix = np.identity(3)
        self._coordinates = np.dot(matrix, self._coordinates)
        
    def translate(self, vector = None):
        """
        Translates an atom according to a translational vector
        """
        if vector is None: vector = np.zeros(3)
        self._coordinates = self._coordinates + vector
        
    def distance(self, atom = None):
        """
        Euclidean distance between two atoms
        """
        if atom is None:
            return np.linalg.norm(self._coordinates - np.zeros(3))
        else:
            return np.linalg.norm(self._coordinates - atom.coordinates)

    def distance_to_point(self, coordinates = None):
        """
        Euclidean distance from an atom to a point
        """
        if coordinates is None:
            return np.linalg.norm(self._coordinates - np.zeros(3))
        else:
            return np.linalg.norm(self._coordinates - coordinates)
    
    #
    # OVERRIDE DEFAULT METHODS
    #
    def __repr__(self):
        return '<{0.__class__.__name__}: [{0.name}, {0.number}]:({0.x:.3f}, {0.y:.3f}, {0.z:.3f})>'.format(self)
    


"""
    Support Library to calculate RMSD difference between two Structure Objects

                                        jbonet @ boliva's lab        2012
"""
import math
import numpy as np


def rigid_transform_3D(A, B):
	A = np.mat(A)
	B = np.mat(B)
	assert len(A) == len(B)
	N = A.shape[0]  # total points
	centroid_A = np.mean(A, axis=0)
	centroid_B = np.mean(B, axis=0)
	AA = A - np.tile(centroid_A, (N, 1))
	BB = B - np.tile(centroid_B, (N, 1))
	H = np.transpose(AA) * BB
	U, S, Vt = np.linalg.svd(H)
	R = Vt.T * U.T
	if np.linalg.det(R) < 0:
		Vt[2, :] *= -1
		R = Vt.T * U.T
	E0   = np.sum( np.sum(np.array(AA) * np.array(AA), axis=0), axis=0) + np.sum( np.sum(np.array(BB) * np.array(BB), axis=0), axis=0)
	RMSD = E0 - (2.0 * np.sum(S))
	RMSD = np.sqrt(np.abs(RMSD / len(np.array(AA))))
	return np.array(R), RMSD



def apply_kabsch_transform(str, Matrix, Vector):
    transtocenter = (str.geometric_center()) * -1
    str.translate_onto_origin()
    str.rotate(Matrix)
    str.translate(-transtocenter)
    str.translate(Vector)


def apply_kabsch_transform2(str, Matrix, Vector, motif_center):
    str.translate(-motif_center)
    str.rotate(Matrix)
    str.translate(motif_center)
    str.translate(Vector)


def superimposition_matrix_vector(structure1, structure2):

    Matrix = kabsch(structure1, structure2)

    Vector = structure2.geometric_center(structure = True, hetero = False, water = False) - \
        structure1.geometric_center(structure = True, hetero = False, water = False)

    for i in range(3):
        for j in range(3):
            if (Matrix[i][j] > 0 and Matrix[i][j] < 1e-5) or (Matrix[i][j] < 0 and Matrix[i][j] > -1e-5):
                Matrix[i][j] = 0
    return Matrix, Vector


def kabsch(str1 = None, str2 = None):
    '''
    Given two structures, kabsch algorithm tries to identify the rotation matrix to move str1 and optimize the RMSD over str2
    '''

    '''
    STEP1: Center into origin
    '''
    centered_str1 = str1.duplicate(hetero = False, water = False, backbone = True)
    centered_str2 = str2.duplicate(hetero = False, water = False, backbone = True)

    centered_str1.translate_onto_origin()
    centered_str2.translate_onto_origin()

    '''
    STEP2: Retrieve all coordinates
    '''
    centered_str1_coordinates = centered_str1.get_backbone_coordinates()
    centered_str2_coordinates = centered_str2.get_backbone_coordinates()

    V, S, Wt = np.linalg.svd(np.dot(np.transpose(centered_str2_coordinates), centered_str1_coordinates))

    reflect = float(str(float(np.linalg.det(V) * np.linalg.det(Wt))))
    if reflect == -1.0:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]

    U = np.dot(V, Wt)

    return U


# TODO: This clearly needs to be improved, but let's leave this for later..
# TODO: refactoring only to CA at the moment
def repositioned_RMSD(str1 = None, str2 = None, c_type = "ca", selection = None):
    '''
    Checking the RMSD between two IDENTICAL structures implies calculating the difference between all their atoms.
    Structures are assumed to have been previously positioned as they are expected to be for the comparison

    This definition accepts that some flexibility might have been applied to the structures
    '''

    '''
    For lack of a better/faster way to check it out, we will consider that two structures are the same if they
    have the same number of atoms.
    Else, we will rise a ValueError
    '''
    if len(str1) != len(str2):
        raise ValueError("Both structures need to be the same for the RMSD of the reposition to be calculated\n")

    '''
    Several flags can be called over this function:
        c_type: the allowed values are:
            all: it will compare the full RMSD
            ca: it will compare the RMSD between the CA atoms
            cb: it will compare the RMSD between the CB atoms -> GLY are ignored
            cb-ca: CB RMSD - CA RMSD -> kind of compares only how the lateral chains have changed
            any other option defaults to 'all'
        selection: an array of positions can be submitted and only those Aa are going to be compared
                   This can be used, for example, if comparing the Interface RMSD
    '''
    allowed_c_type = set(['all', 'ca', 'cb', 'cb-ca'])
    if c_type not in allowed_c_type:
        c_type = 'ca'

    if selection is not None:
        selection_set = set(selection)

    E = 0
    atom_count = 0
    for res in range(len(str1)):
        if selection is None or str1.aminoacids[res].get_residue_num() in selection_set:
            # if c_type == 'all':
            #     for at in range(str1.aminoacids[res].get_residue_size()):
            #         distance = str1.aminoacids[res].get_atoms()[at].get_distance(atom2=str2.aminoacids[res].get_atoms()[at])
            #         atom_count += 1
            #         E += math.pow(distance, 2)
            if c_type == 'ca':
                distance = str1.aminoacids[res].ca.distance(str2.aminoacids[res].ca)
                atom_count += 1
                E += math.pow(distance, 2)
            # if c_type == 'cb' and str1.aminoacids[res].get_cb() is not None:
            #     distance = str1.aminoacids[res].get_cb().get_distance(atom2=str2.aminoacids[res].get_cb())
            #     atom_count += 1
            #     E += math.pow(distance, 2)
            # if c_type == 'cb-ca' and str1.aminoacids[res].get_cb() is not None:
            #     distance_a = str1.aminoacids[res].get_ca().get_distance(atom2=str2.aminoacids[res].get_ca())
            #     distance_b = str1.aminoacids[res].get_cb().get_distance(atom2=str2.aminoacids[res].get_cb())
            #     atom_count += 12
            #     E += math.pow(distance_b - distance_a, 2)

    E = E / atom_count

    return math.sqrt(math.fabs(E))


# def repositioned_RMSD_byMatrix(AV = None, AV2 = None, NormL = None, rotation_matrix1 = None, rotation_matrix2 = None,
#                                translation_vector1 = None, translation_vector2 = None):
#     '''
#     This method will try to implement an RMSD calculation based on the rotation matrices and the translation vectors
#     '''

#     x, y, z, V2 = calculateVandV2(translation_vector1 = translation_vector1, translation_vector2 = translation_vector2)
#     V  = [x, y, z]

#     V1 = calculateV1(rotation_matrix1 = rotation_matrix1, rotation_matrix2 = rotation_matrix2, AV2 = AV2)
#     V3 = calculateV3(V = V, rotation_matrix1 = rotation_matrix1, rotation_matrix2 = rotation_matrix2, AV = AV)

#     '''
#     Calculate RMSD
#     '''
#     rmsd = V2 + NormL - V1 + V3
#     return math.sqrt(math.fabs(rmsd))

# """
#     __Suporting Function to repositioned_RMSD_byMatrix__
# """


# def calculateVandV2(translation_vector1 = None, translation_vector2 = None):
#     '''
#     The translation vector clusters the data in order x, y, z
#     This one creates the vectors:
#         V:    x, y, z
#     And the value V2
#     '''
#     x = translation_vector1[0] - translation_vector2[0]
#     y = translation_vector1[1] - translation_vector2[1]
#     z = translation_vector1[2] - translation_vector2[2]

#     v2 = x * x + y * y + z * z

#     return x, y, z, v2


# def calculateV3(V = None, rotation_matrix1 = None, rotation_matrix2 = None, AV = None):
#     '''
#     This one calculates V3 from the arrays V and AV and the rotation matrices
#         AV[0] : average_x
#         AV[1] : average_y
#         AV[2] : average_z
#         V[0]  : x
#         V[1]  : y
#         V[2]  : z
#     '''

#     v3 = 0

#     v3 = v3 + ((rotation_matrix1[0][0] - rotation_matrix2[0][0]) * AV[0] +
#                (rotation_matrix1[0][1] - rotation_matrix2[0][1]) * AV[1] +
#                (rotation_matrix1[0][2] - rotation_matrix2[0][2]) * AV[2]) * V[0]
#     v3 = v3 + ((rotation_matrix1[1][0] - rotation_matrix2[1][0]) * AV[0] +
#                (rotation_matrix1[1][1] - rotation_matrix2[1][1]) * AV[1] +
#                (rotation_matrix1[1][2] - rotation_matrix2[1][2]) * AV[2]) * V[1]
#     v3 = v3 + ((rotation_matrix1[2][0] - rotation_matrix2[2][0]) * AV[0] +
#                (rotation_matrix1[2][1] - rotation_matrix2[2][1]) * AV[1] +
#                (rotation_matrix1[2][2] - rotation_matrix2[2][2]) * AV[2]) * V[2]

#     return 2 * v3


# def calculateV1(rotation_matrix1 = None, rotation_matrix2 = None, AV2 = None):
#     '''
#     This one calculates V1 from the rotation matrices and the AV2 vector
#     '''

#     v1 = 0

#     v1 = v1 + (rotation_matrix1[0][0] * rotation_matrix2[0][0] +
#                rotation_matrix1[1][0] * rotation_matrix2[1][0] +
#                rotation_matrix1[2][0] * rotation_matrix2[2][0]) * AV2[0]  # xx
#     v1 = v1 + (rotation_matrix1[0][1] * rotation_matrix2[0][1] +
#                rotation_matrix1[1][1] * rotation_matrix2[1][1] +
#                rotation_matrix1[2][1] * rotation_matrix2[2][1]) * AV2[1]  # yy
#     v1 = v1 + (rotation_matrix1[0][2] * rotation_matrix2[0][2] +
#                rotation_matrix1[1][2] * rotation_matrix2[1][2] +
#                rotation_matrix1[2][2] * rotation_matrix2[2][2]) * AV2[2]  # zz

#     v1 = v1 + (rotation_matrix1[0][0] * rotation_matrix2[0][1] +
#                rotation_matrix1[0][1] * rotation_matrix2[0][0] +
#                rotation_matrix1[1][0] * rotation_matrix2[1][1] +
#                rotation_matrix1[1][1] * rotation_matrix2[1][0] +
#                rotation_matrix1[2][0] * rotation_matrix2[2][1] +
#                rotation_matrix1[2][1] * rotation_matrix2[2][0]) * AV2[3]  # xy
#     v1 = v1 + (rotation_matrix1[0][0] * rotation_matrix2[0][2] +
#                rotation_matrix1[0][2] * rotation_matrix2[0][0] +
#                rotation_matrix1[1][0] * rotation_matrix2[1][2] +
#                rotation_matrix1[1][2] * rotation_matrix2[1][0] +
#                rotation_matrix1[2][0] * rotation_matrix2[2][2] +
#                rotation_matrix1[2][2] * rotation_matrix2[2][0]) * AV2[4]  # xz
#     v1 = v1 + (rotation_matrix1[0][1] * rotation_matrix2[0][2] +
#                rotation_matrix1[0][2] * rotation_matrix2[0][1] +
#                rotation_matrix1[1][1] * rotation_matrix2[1][2] +
#                rotation_matrix1[1][2] * rotation_matrix2[1][1] +
#                rotation_matrix1[2][1] * rotation_matrix2[2][2] +
#                rotation_matrix1[2][2] * rotation_matrix2[2][1]) * AV2[5]  # yz

#     return 2 * v1


# def calculateAVandAV2andNormL(str1):
#     '''
#     This one creates the vectors:
#         AV:     average_x, average_y, average_z
#         AV2:    xx, yy, zz, xy, xz, yz
#     and the NormL value
#     '''

#     average_x, average_y, average_z, total_atoms = 0, 0, 0, 0
#     xx, yy, zz, xy, xz, yz = 0, 0, 0, 0, 0, 0

#     for residue in str1.get_residues():
#         for atom in residue.get_atoms():
#             average_x = average_x + atom.get_x()
#             average_y = average_y + atom.get_y()
#             average_z = average_z + atom.get_z()

#             xx = xx + atom.get_x() * atom.get_x()
#             yy = yy + atom.get_y() * atom.get_y()
#             zz = zz + atom.get_z() * atom.get_z()
#             xy = xy + atom.get_x() * atom.get_y()
#             xz = xz + atom.get_x() * atom.get_z()
#             yz = yz + atom.get_y() * atom.get_z()

#             total_atoms += 1

#     average_x = average_x / float(total_atoms)
#     average_y = average_y / float(total_atoms)
#     average_z = average_z / float(total_atoms)

#     xx = xx / float(total_atoms)
#     yy = yy / float(total_atoms)
#     zz = zz / float(total_atoms)
#     xy = xy / float(total_atoms)
#     xz = xz / float(total_atoms)
#     yz = yz / float(total_atoms)

#     NormL = 2 * (xx + yy + zz)

#     return average_x, average_y, average_z, xx, yy, zz, xy, xz, yz, NormL



"""PDBTMlink

author: jbonet
date:   02/2014

@oliva's lab
"""

"""
Import Standard Libraries
"""
import os, re
import urllib.request, urllib.parse, urllib.error

"""
Dependences in SBI library
"""
from SBILib.databases import PDBTMftp
from SBILib.beans.Path     import Path
from SBILib.beans.File     import File

class PDBTMlink(object):

    def __init__(self, local = None):
        self._local     = os.path.abspath(local)
        self._main      = None
        self._pdbtmfile = None
        if local is not None:
            self.local = local

    """ATTRIBUTES"""
    @property
    def local(self):         return self._local
    @local.setter
    def local(self, value):
        self._local     = os.path.abspath(value)
        self._pdbtmfile = os.path.join(self._local, 'pdbtm.gz')

    @property
    def source(self):
        return PDBTMftp['show']

    """BOOLEANS"""
    @property
    def has_local(self):    return self._local is not None

    """METHODS"""
    def download(self):
        if not self.has_local:
            raise NameError('A local drugBank database directory must be defined.')

        Path.mkdir(self.local)
        here = os.getcwd()
        os.chdir(self.local)
        os.system("svn export {0}".format(PDBTMftp['svn']))

        self._process()

        return True

    @property
    def localTM(self):
        tmoFile = File(self._pdbtmfile, 'r')
        for tm_line in tmoFile.descriptor:
            yield tm_line

    """PRIVATE METHODS"""
    def _process(self):
        tmoFile = File(self._pdbtmfile,'w', True)
        for xmlfile in Path.list_files(os.path.join(self._local,'pdbtm/database/'), '*.xml'):
            xmldata = TM(pdb = os.path.splitext(os.path.split(xmlfile)[1])[0].upper())
            skip_chains = set()
            read = False
            fdxml = open(xmlfile)
            for line in fdxml:
                if line.startswith('    <TMRES>'):     xmldata.tmres  = line
                elif line.startswith('    <TMTYPE'):   xmldata.tmtype = line
                elif line.startswith('    <PDBKWRES'): xmldata.kwres  = line
                elif line.startswith('  <SIDEDEFINITION'):
                    m = re.search('Side1="(\S+)"', line)
                    xmldata.side = m.group(1)
                elif line.startswith('      <APPLY_TO_CHAIN'):
                    m = re.search('NEW_CHAINID=\"(\S{1})\"', line)
                    if m: skip_chains.add(m.group(1))
                elif line.startswith('  <CHAIN '):
                    m = re.search('CHAINID=\"(\S{1})\" NUM_TM=\"(\d{1})\" TYPE=\"(\S+)\"', line)
                    if m:
                        chain, num, tmtype = m.group(1), m.group(2), m.group(3)
                        if not chain in skip_chains:
                            cdata = tuple([chain, num, tmtype])
                            xmldata.set_chain(cdata)
                            read = True
                elif line.startswith('    <REGION ') and read:
                    m = re.search('pdb_beg=\"(\-*\d+\w*)\"[\s\S]+pdb_end=\"(\-*\d+\w*)\"\s+type=\"(\w{1})\"', line)
                    ini, end, tmtype = m.group(1), m.group(2), m.group(3)
                    xmldata.set_chain(cdata, tuple([ini, end, tmtype]))
                elif line.startswith('  </CHAIN>'): read = False
            fdxml.close()
            if len(xmldata.chains) > 0:
                tmoFile.write(str(xmldata)+"\n")
        tmoFile.close()


class TM(object):
    section_types = {'1':'Side1', '2':'Side2', 'i':'Inside', 'o':'Outside',
                     'B':'Beta-strand', 'H':'alpha-helix', 'C':'coil',
                     'I':'membrane-inside', 'L':'membrane-loop', 'F':'interfacial helix', 'U':'unknown localizations'}

    def __init__(self, pdb= None, inline = None):
        if inline is not None: inline = inline.strip().split('\t')

        self._pdb    = pdb  if inline is None else eval(inline[0])
        self._tmres  = ''   if inline is None else eval(inline[1])
        self._tmtype = ''   if inline is None else eval(inline[2])
        self._kwres  = ''   if inline is None else eval(inline[3])
        self._side   = ''   if inline is None else eval(inline[4])
        self._chains = {}   if inline is None else eval(inline[5])

    @property
    def pdb(self): return self._pdb
    @property
    def tmres(self): return self._tmres
    @tmres.setter
    def tmres(self, value): self._tmres = TM._clean_xml(value)
    @property
    def tmtype(self): return self._tmtype
    @tmtype.setter
    def tmtype(self, value): self._tmtype = TM._clean_xml(value)
    @property
    def kwres(self): return self._kwres
    @kwres.setter
    def kwres(self, value): self._kwres = TM._clean_xml(value)
    @property
    def side(self): return self._side
    @side.setter
    def side(self, value): self._side = value
    @property
    def chains(self): return self._chains

    def set_chain(self, key, value = None):
        if value is None:
            self._chains.setdefault(key,[])
        else:
            self._chains[key].append(value)

    @staticmethod
    def _clean_xml(xml):
        return re.search(">([\S\s]+)</", xml).group(1)

    def __str__(self):
        data = []
        data.extend([self._pdb,self._tmres,self._tmtype,self._kwres, self._side])
        data.append(self._chains)
        return "\t".join([repr(x) for x in data])


